import math
import wx
import wx.aui
import wx.py.dispatcher as dp
import numpy
import matplotlib
matplotlib.use('module://bsmedit.bsm.bsmbackend')
from matplotlib.figure import Figure
from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg as FigureCanvas
from matplotlib.backends.backend_wx import NavigationToolbar2Wx as NavigationToolbar
from matplotlib.backends.backend_wx import FigureManagerWx
from matplotlib._pylab_helpers import Gcf
import matplotlib.pyplot as plt
from matplotlib import rcParams
from .bsmxpm import home_xpm, back_xpm, forward_xpm, pan_xpm, zoom_xpm, \
                    cursor_xpm, save_xpm, copy_xpm
from .. import to_byte
rcParams.update({'figure.autolayout': True})
rcParams.update({'toolbar': 'None'})
matplotlib.interactive(True)


class DataCursor(object):
    xoffset, yoffset = -20, 20
    text_template = 'x: %0.2f\ny: %0.2f'

    def __init__(self):
        self.annotations = []
        self.lines = []
        self.enable = False
        self.active = None
        self.mx, self.my = None, None
        self.pickEvent = False

    def __call__(self, event):
        if not self.enable:
            return
        if not event.mouseevent.xdata or not event.mouseevent.ydata:
            return
        if not event.artist:
            return
        xm, ym = event.mouseevent.x, event.mouseevent.y
        if self.get_annotation(xm, ym) is not None:
            # click in the box of existing annotation, ignore it
            return

        line = event.artist
        if self.active and self.active.get_visible():
            # Check whether the axes of active annotation is same as line,
            # which may happen in a figure with subplots. If not, create one
            # with the axes of line
            if self.active.axes != line.axes:
                self.active = None
        if self.active is None:
            self.create_annotation(line)
        idx = self.annotations.index(self.active)
        self.lines[idx] = line

        # find the closest point on the line
        x, y = line.get_xdata(), line.get_ydata()
        xc, yc = event.mouseevent.xdata, event.mouseevent.ydata
        idx = (numpy.square(x - xc) + numpy.square(y - yc)).argmin()
        xn, yn = x[idx], y[idx]
        if xn is not None:
            self.active.xy = xn, yn
            self.active.set_text(self.text_template % (xn, yn))
            self.active.set_visible(True)
            event.canvas.draw()
        self.pickEvent = True

    def keyboard_move(self, left, step=1):
        if not self.active:
            return
        idx = self.annotations.index(self.active)
        line = self.lines[idx]
        x, y = line.get_xdata(), line.get_ydata()
        xc, yc = self.active.xy
        idx = (numpy.square(x - xc)).argmin()
        idx_new = idx
        if left:
            idx_new -= step
        else:
            idx_new += step
        idx_new = min(len(x)-1, idx_new)
        idx_new = max(0, idx_new)
        if idx == idx_new:
            return
        xn, yn = x[idx_new], y[idx_new]
        if xn is not None:
            self.active.xy = xn, yn
            self.active.set_text(self.text_template % (xn, yn))
            self.active.set_visible(True)

    def set_enable(self, enable):
        self.enable = enable
        if self.active:
            if enable:
                self.active.get_bbox_patch().set_facecolor('yellow')
            else:
                self.active.get_bbox_patch().set_facecolor('white')

    def mouse_move(self, x, y):
        """move the annotation position"""
        # return if no active annotation or the mouse is not pressed
        if self.mx is None or self.my is None or self.active is None:
            return False
        # re-position the active annotation based on the mouse movement
        dx = x - self.mx
        dy = y - self.my
        dis = math.sqrt(dx**2 + dy**2)
        if dis > 50:
            (px, py) = (-20, 20)
            if dx > 80:
                px = 60
            elif dx > 40:
                px = 20
            if dy < -80:
                py = -60
            elif dy < -40:
                py = -20
            self.active.xyann = (px, py)
            return True
        return False

    def get_annotation(self, x, y):
        for ant in self.annotations:
            box = ant.get_bbox_patch().get_extents()
            if box.contains(x, y):
                return ant
        return None

    def mouse_pressed(self, x, y):
        """
        select the active annotation which is closest to the mouse position
        """
        # ignore the event triggered immediately after pick_event
        if self.pickEvent:
            self.pickEvent = False
            return False
        # just created the new annotation, do not move to others
        if self.active and (not self.active.get_visible()):
            return False
        # search the closest annotation
        self.mx, self.my = x, y
        active = self.get_annotation(x, y)
        self.set_active(active)
        # return True for the parent to redraw
        return True

    def mouse_released(self, x, y):
        """release the mouse"""
        self.mx, self.my = None, None

    def get_active(self):
        """retrieve the active annotation"""
        return self.active

    def set_active(self, ant):
        """set the active annotation"""
        if ant and (ant not in self.annotations):
            return False
        if self.active == ant:
            return True
        if self.active:
            self.active.get_bbox_patch().set_facecolor('white')
        self.active = ant
        if self.active:
            self.active.get_bbox_patch().set_facecolor('yellow')
        return True

    def get_annotations(self):
        """return all the annotations"""
        return self.annotations

    def create_annotation(self, line):
        """create the annotation and set it active"""
        ant = line.axes.annotate(self.text_template,
                          xy=(0, 0),
                          xytext=(self.xoffset, self.yoffset),
                          textcoords='offset points',
                          ha='right',
                          va='bottom',
                          bbox=dict(boxstyle='round,pad=0.5',
                                    fc='yellow',
                                    alpha=0.5),
                          arrowprops=dict(arrowstyle='->',
                                          connectionstyle='arc3,rad=0'))
        ant.set_visible(False)
        self.annotations.append(ant)
        self.lines.append(line)
        self.set_active(ant)

    def ProcessCommand(self, cmd):
        """process the context menu command"""
        if cmd == MatplotPanel.clsID_delete_datatip:
            if not self.active:
                return False
            idx = self.annotations.index(self.active)
            self.active.remove()
            del self.annotations[idx]
            del self.lines[idx]
            self.active = None
            return True
        elif cmd == MatplotPanel.clsID_clear_datatip:
            for ant in self.annotations:
                try:
                    # the call may fail. For example,
                    # 1) create a figure and plot some curve
                    # 2) create a datatip
                    # 3) call clf() to clear the figure, the datatip will be
                    #    cleared, but we will not know
                    ant.remove()
                except:
                    pass
            self.annotations = []
            self.lines = []
            self.active = None
            return True
        return False


class Toolbar(NavigationToolbar):
    def __init__(self, canvas, figure):
        if matplotlib.__version__ < '3.3.0':
            self._init_toolbar = self.init_toolbar
        else:
            self._init_toolbar = self.init_toolbar_empty
        NavigationToolbar.__init__(self, canvas)

        if matplotlib.__version__ >= '3.3.0':
            self.init_toolbar()
        self.SetWindowStyle(wx.TB_HORIZONTAL | wx.TB_FLAT)
        self.figure = figure
        self.datacursor = DataCursor()
        self.canvas.mpl_connect('pick_event', self.datacursor)
        self.canvas.mpl_connect('motion_notify_event', self.OnMove)
        self.canvas.mpl_connect('button_press_event', self.OnPressed)
        self.canvas.mpl_connect('button_release_event', self.OnReleased)
        self.canvas.mpl_connect('scroll_event', self.OnZoomFun)
        # clear the view history
        wx.CallAfter(self._nav_stack.clear)

    def OnPressed(self, event):
        if self.mode != 'datatip':
            return
        # some lines may be added
        for l in self.figure.gca().lines:
            l.set_picker(5)
        if self.datacursor.mouse_pressed(event.x, event.y):
            self.canvas.draw()

    def OnReleased(self, event):
        if self.mode != 'datatip':
            return
        self.datacursor.mouse_released(event.x, event.y)

    def OnMove(self, event):
        if self.mode != 'datatip':
            return
        if self.datacursor.mouse_move(event.x, event.y):
            self.canvas.draw()

    def OnZoomFun(self, event):
        # get the current x and y limits
        if not self.GetToolState(self.wx_ids['Zoom']):
            return
        if self._nav_stack.empty():
            self.push_current()
        base_scale = 2.0
        ax = self.figure.gca()
        cur_xlim = ax.get_xlim()
        cur_ylim = ax.get_ylim()

        xdata = event.xdata  # get event x location
        ydata = event.ydata  # get event y location
        if xdata is None:
            return
        if ydata is None:
            return

        if event.button == 'down':
            # deal with zoom in
            scale_factor = 1.0 / base_scale
        elif event.button == 'up':
            # deal with zoom out
            scale_factor = base_scale
        else:
            # deal with something that should never happen
            scale_factor = 1.0

        new_width = (cur_xlim[1] - cur_xlim[0]) * scale_factor
        new_height = (cur_ylim[1] - cur_ylim[0]) * scale_factor

        relx = (cur_xlim[1] - xdata) / (cur_xlim[1] - cur_xlim[0])
        rely = (cur_ylim[1] - ydata) / (cur_ylim[1] - cur_ylim[0])
        xzoom = yzoom = True
        if wx.GetKeyState(wx.WXK_CONTROL_X):
            yzoom = False
        elif wx.GetKeyState(wx.WXK_CONTROL_Y):
            xzoom = False
        if (xzoom) and new_width * (1 - relx) > 0:
            ax.set_xlim(
                [xdata - new_width * (1 - relx), xdata + new_width * (relx)])
        if (yzoom) and new_height * (1 - rely) > 0:
            ax.set_ylim(
                [ydata - new_height * (1 - rely), ydata + new_height * (rely)])
        self.canvas.draw()

    def init_toolbar_empty(self):
        # deprecated in 3.3.0
        pass
    def init_toolbar(self):
        toolitems = (
            ('Home', 'Reset original view', home_xpm, 'home'),
            ('Back', 'Back to  previous view', back_xpm, 'back'),
            ('Forward', 'Forward to next view', forward_xpm, 'forward'),
            (None, None, None, None),
            ('Pan', 'Pan axes with left mouse, zoom with right', pan_xpm,
             'pan'),
            ('Zoom', 'Zoom to rectangle', zoom_xpm, 'zoom'),
            ('Datatip', 'Show the data tip', cursor_xpm, 'datatip'),
            (None, None, None, None),
            ('Save', 'Save the figure', save_xpm, 'save_figure'),
            ('Copy', 'Copy to clipboard', copy_xpm, 'copy_figure'),
            #(None, None, None, None),
            #('Print', 'Print the figure', print_xpm, 'print_figure'),
        )

        self._parent = self.canvas.GetParent()
        self.ClearTools()
        self.wx_ids = {}
        self.SetToolBitmapSize((16, 16))
        for (text, tooltip_text, image_file, callback) in toolitems:
            if text is None:
                self.AddSeparator()
                continue
            self.wx_ids[text] = wx.NewId()
            if text in ['Pan', 'Zoom', 'Datatip']:
                self.AddCheckTool(self.wx_ids[text],
                                  text,
                                  wx.Bitmap(to_byte(image_file)),
                                  shortHelp=text,
                                  longHelp=tooltip_text)
            else:
                self.AddTool(self.wx_ids[text], text,
                             wx.Bitmap(to_byte(image_file)),
                             kind=wx.ITEM_NORMAL, shortHelp=tooltip_text)
            self.Bind(wx.EVT_TOOL,
                      getattr(self, callback),
                      id=self.wx_ids[text])

        self.Realize()

    def copy_figure(self, evt):
        # self.canvas.Copy_to_Clipboard(event=evt)
        bmp_obj = wx.BitmapDataObject()
        bmp_obj.SetBitmap(self.canvas.bitmap)

        if not wx.TheClipboard.IsOpened():
            open_success = wx.TheClipboard.Open()
            if open_success:
                wx.TheClipboard.SetData(bmp_obj)
                wx.TheClipboard.Flush()
                wx.TheClipboard.Close()

    def print_figure(self, evt):
        self.canvas.Printer_Print(event=evt)

    def zoom(self, *args):
        """activate the zoom mode"""
        self.ToggleTool(self.wx_ids['Datatip'], False)
        super(Toolbar, self).zoom(*args)

    def pan(self, *args):
        """activated the pan mode"""
        self.ToggleTool(self.wx_ids['Datatip'], False)
        super(Toolbar, self).pan(*args)

    def datatip(self, evt):
        """activate the datatip mode"""
        # disable the pan/zoom mode
        self.ToggleTool(self.wx_ids['Zoom'], False)
        self.ToggleTool(self.wx_ids['Pan'], False)
        self._active = None
        if hasattr(self, '_idPress'):
            self._idPress = self.canvas.mpl_disconnect(self._idPress)

        if hasattr(self, '_idRelease'):
            self._idRelease = self.canvas.mpl_disconnect(self._idRelease)
        self.canvas.widgetlock.release(self)
        for a in self.canvas.figure.get_axes():
            a.set_navigate_mode(self._active)

        for l in self.figure.gca().lines:
            l.set_picker(5)
        self.mode = "datatip"
        self.set_message(self.mode)
        self.datacursor.set_enable(evt.GetInt())

    def set_message(self, s):
        """show the status message"""
        dp.send(signal='frame.show_status_text', text=s, index=1, width=160)


class MatplotPanel(wx.Panel):
    clsFrame = None
    clsID_new_figure = wx.NOT_FOUND
    isInitialized = False
    kwargs = {}
    clsID_delete_datatip = wx.NewId()
    clsID_clear_datatip = wx.NewId()

    def __init__(self, parent, title=None, num=-1, thisFig=None):
        # set the size to positive value, otherwise the toolbar will assert
        # wxpython/ext/wxWidgets/src/gtk/bitmap.cpp(539): assert ""width > 0 &&
        # height > 0"" failed in Create(): invalid bitmap size
        wx.Panel.__init__(self, parent, size=(100, 100))
        # initialize matplotlib stuff
        self.figure = thisFig
        if not self.figure:
            self.figure = Figure(None, None)
        self.canvas = FigureCanvas(self, -1, self.figure)
        # since matplotlib 3.2, it does not allow canvas size to become smaller
        # than MinSize in wx backend. So the canvas size (e.g., (640, 480))may
        # be large than the window size.
        self.canvas.SetMinSize((1, 1))
        #self.canvas.manager = self

        self.num = num
        if title is None:
            title = 'Figure %d' % self.num
        self.title = title
        self.isdestory = False
        szAll = wx.BoxSizer(wx.VERTICAL)

        self.figure.set_label(title)
        self.toolbar = Toolbar(self.canvas, self.figure)
        szAll.Add(self.toolbar, 0, wx.EXPAND)
        szAll.Add(self.canvas, 1, wx.LEFT | wx.TOP | wx.GROW)

        self.toolbar.update()
        self.SetSizer(szAll)


        self.figmgr = FigureManagerWx(self.canvas, num, self)
        self.Bind(wx.EVT_CLOSE, self._onClose)

        self.canvas.mpl_connect('button_press_event', self._onClick)
        dp.connect(self.simLoad, 'sim.loaded')
        self.Bind(wx.EVT_CONTEXT_MENU, self.OnContextMenu)
        self.Bind(wx.EVT_MENU, self.OnProcessCommand, id=self.clsID_delete_datatip)
        self.Bind(wx.EVT_MENU, self.OnProcessCommand, id=self.clsID_clear_datatip)
        self.Bind(wx.EVT_MENU, self.OnProcessCommand, id=wx.ID_NEW)
        self.Bind(wx.EVT_CHAR_HOOK, self.OnKeyDown)

    def GetToolBar(self):
        """Override wxFrame::GetToolBar as we don't have managed toolbar"""
        return self.toolbar

    def simLoad(self, num):
        for l in self.figure.gca().lines:
            if hasattr(l, 'trace'):
                sz = len(l.get_ydata())
                for s in l.trace:
                    if (not s) or (not s.startswith(str(num) + '.')):
                        continue
                    #dispatcher.send(signal='sim.trace_buf', objects=s, size=sz)

    def _onClick(self, event):
        if event.dblclick:
            self.toolbar.home()

    def OnKeyDown(self, evt):
        if self.toolbar.mode != 'datatip':
            return
        keycode = evt.GetKeyCode()
        step = 1
        if evt.ShiftDown():
            step = 10
        if keycode == wx.WXK_LEFT:
            self.toolbar.datacursor.keyboard_move(True, step=step)
        elif keycode == wx.WXK_RIGHT:
            self.toolbar.datacursor.keyboard_move(False, step=step)
        else:
            evt.Skip()

    def OnProcessCommand(self, evt):
        if self.toolbar.datacursor.ProcessCommand(evt.GetId()):
            self.canvas.draw()

    def _show_context_menu(self):
        if self.toolbar.mode != 'datatip':
            return

        menu = wx.Menu()
        delMenu = menu.Append(self.clsID_delete_datatip, "Delete current datatip")
        note = self.toolbar.datacursor.active
        delMenu.Enable((note is not None) and note.get_visible())
        menu.Append(self.clsID_clear_datatip, "Delete all datatips")
        self.PopupMenu(menu)
        menu.Destroy()

    def OnContextMenu(self, event):
        # Show menu after the current and pending event handlers have been
        # completed, otherwise it causes the following error in some system
        # (e.g., xubuntu, matplotlib 3.2.2, wx 4.1.0), and the menu doesn't show.
        # GLib-GObject-CRITICAL **: g_object_set_data: assertion 'G_IS_OBJECT
        # (object)' failed
        wx.CallAfter(self._show_context_menu)

    def _onClose(self, evt):
        self.canvas.close_event()
        self.canvas.stop_event_loop()
        Gcf.destroy(self.num)

    def destroy(self, *args):
        if self.isdestory is False:
            dp.send('frame.delete_panel', panel=self)
            wx.WakeUpIdle()

    def Destroy(self, *args, **kwargs):
        self.isdestory = True
        self.canvas.close_event()
        self.canvas.stop_event_loop()
        Gcf.destroy(self.num)
        return super(MatplotPanel, self).Destroy(*args, **kwargs)

    def GetTitle(self):
        """return the figure title"""
        return self.title

    def SetTitle(self, title):
        """set the figure title"""
        if title == self.title:
            return
        self.title = title
        dp.send('frame.update_panel_title', pane=self, title=self.title)

    def show(self):
        """show figure"""
        if self.IsShown() is False:
            self.canvas.draw()
            dp.send('frame.show_panel', panel=self)

    def update_buffer(self, bufs):
        """update the data used in plot_trace"""
        for l in self.figure.gca().lines:
            if hasattr(l, 'trace'):
                x = l.trace[0]
                y = l.trace[1]
                if x is None:
                    if y in bufs:
                        l.set_data(numpy.arange(len(bufs[y])), bufs[y])
                elif x in bufs or y in bufs:
                    xd = l.get_xdata()
                    yd = l.get_ydata()
                    if y in bufs:
                        yd = bufs[y]
                    if x in bufs:
                        xd = bufs[x]
                    if len(xd) != len(yd):
                        sz = min(len(xd), len(yd))
                        xd = xd[0:sz]
                        yd = yd[0:sz]
                    l.set_data(xd, yd)
                if hasattr(l, 'autorelim') and l.autorelim:
                    #Need both of these in order to rescale
                    self.figure.gca().relim()
                    self.figure.gca().autoscale_view()
        self.canvas.draw()

    def plot_trace(self, x, y, autorelim, *args, **kwargs):
        """plot and trace"""
        if y is None:
            return
        if x is None:
            l, = self.figure.gca().plot(list(y.values())[0], *args, **kwargs)
            l.trace = [None, list(y.keys())[0]]
        else:
            xd = list(x.values())[0]
            yd = list(y.values())[0]
            if len(xd) != len(yd):
                sz = min(len(xd), len(yd))
                if sz > 0:
                    xd = xd[0:sz]
                    yd = yd[0:sz]
                else:
                    xd = 0
                    yd = 0
            l, = self.figure.gca().plot(xd, yd, *args, **kwargs)
            l.trace = [list(x.keys())[0], list(y.keys())[0]]
        l.autorelim = autorelim
        self.canvas.draw()
    def set_window_title(self, label):
        pass

    @classmethod
    def setactive(cls, pane):
        """set the active figure"""
        if pane and isinstance(pane, MatplotPanel):
            Gcf.set_active(pane)

    @classmethod
    def addFigure(cls, title=None, num=None, thisFig=None):
        direction = cls.kwargs.get('direction', 'top')
        fig = cls(cls.clsFrame, title=title, num=num, thisFig=thisFig)
        # set the minsize to be large enough to avoid some following assert; it
        # will not eliminate all as if a page is added to a notebook, the
        # minsize of notebook is not the max of all its children pages (check
        # frameplus.py).
        # wxpython/ext/wxWidgets/src/gtk/bitmap.cpp(539): assert ""width > 0 &&
        # height > 0"" failed in Create(): invalid bitmap size
        dp.send('frame.add_panel',
                panel=fig,
                direction=direction,
                title=fig.GetTitle(),
                target=Gcf.get_active(),
                minsize=(75, 75))
        return fig

    @classmethod
    def Initialize(cls, frame, **kwargs):
        if cls.isInitialized:
            return
        cls.isInitialized = True
        cls.clsFrame = frame
        cls.kwargs = kwargs
        resp = dp.send('frame.add_menu',
                       path='File:New:Figure',
                       rxsignal='bsm.figure')
        if resp:
            cls.clsID_new_figure = resp[0][1]

        if cls.clsID_new_figure is not wx.NOT_FOUND:
            dp.connect(cls.ProcessCommand, 'bsm.figure')
        dp.connect(cls.Uninitialize, 'frame.exit')
        dp.connect(cls.Initialized, 'frame.initialized')
        dp.connect(cls.setactive, 'frame.activate_panel')
        dp.connect(cls.OnBufferChanged, 'sim.buffer_changed')

    @classmethod
    def Initialized(cls):
        dp.send('shell.run',
                command='from matplotlib.pyplot import *',
                prompt=False,
                verbose=False,
                history=False)

    @classmethod
    def OnBufferChanged(cls, bufs):
        """the buffer has be changes, update the plot_trace"""
        for p in Gcf.get_all_fig_managers():
            p.update_buffer(bufs)

    @classmethod
    def Uninitialize(cls):
        """destroy the module"""
        Gcf.destroy_all()

    @classmethod
    def ProcessCommand(cls, command):
        """process the menu command"""
        if command == cls.clsID_new_figure:
            plt.figure()


def bsm_initialize(frame, **kwargs):
    """module initialization"""
    MatplotPanel.Initialize(frame, **kwargs)



# PyCParser caching logic
# by Albert Zeyer, 2011
# code under BSD 2-Clause License

# idea:
#   for each parsed file:
#     keep list of which macros have been used, i.e.
#        the dependency list of macros.
#     keep list of all C-stuff which has been added.
#     check last change time of file and all other files we open from here
#        and also store this list.
#     save all.
#  when opening a new file, macro-dependencies, check the last change time
#     of all files and if everything matches, use the cache.

import sys
import os, os.path
if sys.version_info.major == 2:
    import cparser
    from cparser_utils import *
else:
    from . import cparser
    from .cparser_utils import *
import types

# Note: It might make sense to make this somehow configureable.
# However, for now, I'd like to keep things as simple as possible.
# Using /tmp or (a bit better) /var/tmp might have been another
# possibility. However, it makes sense to keep this more permanent
# because when compiling a lot, it can be very time-critical if
# we just remove all the data.
# If wasted space becomes an issue, it is easy to write a script
# which would remove all old/obsolete entries from the cache.
# It makes sense also to keep this global for the whole system
# because the caching system should be able to handle this
# and it should thus only improve the performance.
# It is saved though in the user directory because most probably
# we wouldn't have write permission otherwise.
CACHING_DIR = os.path.expanduser("~/.cparser_caching/")

def sha1(obj):
    import hashlib
    h = hashlib.sha1()
    if sys.version_info.major == 2:
        def h_update(s): h.update(s)
    else:
        def h_update(s):
            if isinstance(s, str): h.update(s.encode("utf-8"))
            else: h.update(s)
    if isinstance(obj, unicode):
        h_update(obj.encode("utf-8"))
    elif isinstance(obj, str):
        h_update(obj)
    elif isinstance(obj, dict):
        h_update("{")
        for k,v in sorted(obj.items()):
            h_update(sha1(k))
            h_update(":")
            h_update(sha1(v))
            h_update(",")
        h_update("}")
    elif isinstance(obj, (list,tuple)):
        h_update("[")
        for v in sorted(obj):
            h_update(sha1(v))
            h_update(",")
        h_update("]")
    else:
        h_update(str(obj))
    return h.hexdigest()

class MyDict(dict):
    def __setattr__(self, key, value):
        assert isinstance(key, (str,unicode))
        self[key] = value
    def __getattr__(self, key):
        try: return self[key]
        except KeyError: raise AttributeError
    def __repr__(self): return "MyDict(" + dict.__repr__(self) + ")"
    def __str__(self): return "MyDict(" + dict.__str__(self) + ")"

class DbObj:
    @classmethod
    def GetFilePath(cls, key):
        h = sha1(key)
        prefix = CACHING_DIR + cls.Namespace
        return prefix + "/" + h[:2] + "/" + h[2:]
    @classmethod
    def Load(cls, key, create=False):
        fn = cls.GetFilePath(key)
        try: f = open(fn, "b")
        except:
            if create:
                obj = cls()
                obj.__dict__["_key"] = key
                return obj
            else:
                return None
        import pickle
        obj = pickle.load(f)
        f.close()
        return obj
    @classmethod
    def Delete(cls, key):
        fn = cls.GetFilePath(key)
        os.remove(fn)
    def delete(self): self.Delete(self._key)
    def save(self):
        fn = self.GetFilePath(self._key)
        try: os.makedirs(os.path.dirname(fn))
        except: pass # ignore file-exists or other errors
        f = open(fn, "wb")
        import pickle
        pickle.dump(self, f)
        f.close()

def getLastChangeUnixTime(filename):
    import os.path
    return os.path.getmtime(filename)

class FileCacheRef(MyDict):
    @classmethod
    def FromCacheData(cls, cache_data):
        ref = cls()
        ref.filedepslist = map(lambda fn: (fn,getLastChangeUnixTime(fn)), cache_data.filenames)
        ref.macros = {}
        for m in cache_data.macroAccessSet:
            ref.macros[m] = cache_data.oldMacros[m]
        return ref
    def match(self, stateStruct):
        for macro in self.macros:
            if stateStruct.macros[macro] != self.macros[macro]:
                return False
        return True
    def checkFileDepListUpToDate(self):
        for fn,unixtime in self.filedepslist:
            if getLastChangeUnixTime(fn) > unixtime:
                return False
        return True

class FileCacheRefs(DbObj, list):
    Namespace = "file-cache-refs"

class FileCache(DbObj, MyDict):
    Namespace = "file-cache"
    @classmethod
    def FromCacheData(cls, cache_data, key):
        obj = cls()
        obj.__dict__["_key"] = key
        obj.additions = cache_data.additions
        return obj
    def apply(self, stateStruct):
        for k,l in self.additions.items():
            a = getattr(stateStruct, k)
            if isinstance(a, (list,StateListWrapper)):
                a.extend(l)
            elif isinstance(a, (dict,StateDictWrapper)):
                for dk,dv in l:
                    if dv is None:
                        a.pop(dk)
                    else:
                        a[dk] = dv
            else:
                assert False, "unknown attribute " + k + ": " + str(a)

def check_cache(stateStruct, full_filename):
    filecaches = FileCacheRefs.Load(full_filename)
    if filecaches is None: return None

    for filecacheref in filecaches:
        if not filecacheref.match(stateStruct):
            continue
        if not filecacheref.checkFileDepListUpToDate():
            FileCache.Delete(filecacheref)
            filecaches.remove(filecacheref)
            filecaches.save()
            return None
        filecache = FileCache.Load(filecacheref)
        assert filecache is not None, sha1(filecacheref) + " not found in " + FileCache.Namespace
        return filecache

    return None

def save_cache(cache_data, full_filename):
    filecaches = FileCacheRefs.Load(full_filename, create=True)
    filecacheref = FileCacheRef.FromCacheData(cache_data)
    filecaches.append(filecacheref)
    filecaches.save()
    filecache = FileCache.FromCacheData(cache_data, key=filecacheref)
    filecache.save()

# Note: This does more than State.preprocess. In case it hits a cache,
# it applies all effects up to cpre3 and ignores the preprocessing.
# Note also: This is a generator. In the cache hit case, it yields nothing.
# Otherwise, it doesn't do any further processing and it just yields the rest.
def State__cached_preprocess(stateStruct, reader, full_filename, filename):
    if not full_filename:
        # shortcut. we cannot use caching if we don't have the full filename.
        for c in generic_class_method(cparser.State.preprocess)(stateStruct, reader, full_filename, filename):
            yield c
        return

    if stateStruct._cpre3_atBaseLevel:
        try:
            cached_entry = check_cache(stateStruct, full_filename)
            if cached_entry is not None:
                cached_entry.apply(stateStruct)
                return
        except Exception as e:
            print("(Safe to ignore) Error while reading C parser cache for %s : %s" % (filename, str(e)))
            # Try to delete old references if possible. Otherwise we might always hit this.
            try: FileCacheRefs.Delete(full_filename)
            except: pass

    assert isinstance(stateStruct, StateWrapper)
    stateStruct.cache_pushLevel()
    stateStruct._filenames.add(full_filename)
    for c in generic_class_method(cparser.State.preprocess)(stateStruct, reader, full_filename, filename):
        yield c
    cache_data = stateStruct.cache_popLevel()

    save_cache(cache_data, full_filename)

class StateDictWrapper:
    def __init__(self, d, addList, addSet=None, accessSet=None):
        self._addList = addList
        self._addSet = addSet
        self._accessSet = accessSet
        self._dict = d
    def __getattr__(self, k):
        return getattr(self._dict, k)
    def __setitem__(self, k, v):
        assert v is not None
        self._dict[k] = v
        self._addList.append((k,v))
        if self._addSet is not None:
            self._addSet.add(k)
    def __getitem__(self, k):
        if self._accessSet is not None:
            assert self._addSet is not None
            if not k in self._addSet: # we only care about it if we didn't add it ourself
                self._accessSet.add(k)
        return self._dict[k]
    def __contains__(self, k): return self.has_key(k)
    def has_key(self, k):
        haskey = self._dict.__contains__(k)
        if haskey and self._accessSet is not None:
            assert self._addSet is not None
            if not k in self._addSet: # we only care about it if we didn't add it ourself
                self._accessSet.add(k)
        return haskey
    def pop(self, k):
        self._dict.pop(k)
        self._addList.append((k,None))
        if self._addSet is not None:
            self._addSet.discard(k)
    def __repr__(self): return "StateDictWrapper(" + repr(self._dict) + ")"
    def __str__(self): return "StateDictWrapper(" + str(self._dict) + ")"

class StateListWrapper:
    def __init__(self, l, addList):
        self._addList = addList
        self._list = l
    def __getattr__(self, k):
        return getattr(self._list, k)
    def __iadd__(self, l):
        self._list.extend(l)
        self._addList.extend(l)
        return self
    def append(self, v):
        self._list.append(v)
        self._addList.append(v)
    def extend(self, l):
        self._list.extend(l)
        self._addList.extend(l)
    def __repr__(self): return "StateListWrapper(" + repr(self._list) + ")"
    def __str__(self): return "StateListWrapper(" + str(self._list) + ")"

class StateWrapper:
    WrappedDicts = ("macros","typedefs","structs","unions","enums","funcs","vars","enumconsts")
    WrappedLists = ("contentlist",)
    LocalAttribs = ("_stateStruct", "_cache_stack", "_additions", "_macroAccessSet", "_macroAddSet", "_filenames", "_cpre3_atBaseLevel")
    def __init__(self, stateStruct):
        self._stateStruct = stateStruct
        self._cache_stack = []
    def __getattr__(self, k):
        if k in self.LocalAttribs: raise AttributeError # normally we shouldn't get here but just in case
        if len(self._cache_stack) > 0:
            if k in self.WrappedDicts:
                kwattr = {'d': getattr(self._stateStruct, k), 'addList': self._additions[k]}
                if k == "macros":
                    kwattr["accessSet"] = self._macroAccessSet
                    kwattr["addSet"] = self._macroAddSet
                return StateDictWrapper(**kwattr)
            if k in self.WrappedLists:
                return StateListWrapper(getattr(self._stateStruct, k), addList=self._additions[k])
        attr = getattr(self._stateStruct, k)
        if isinstance(attr, types.MethodType):
            attr = rebound_instance_method(attr, self)
        return attr
    def __repr__(self):
        return "<StateWrapper of " + repr(self._stateStruct) + ">"
    def __setattr__(self, k, v):
        if k in self.LocalAttribs:
            self.__dict__[k] = v
            return
        if k in self.WrappedLists and isinstance(v, StateListWrapper): return # ignore. probably iadd or so.
        setattr(self._stateStruct, k, v)
    def cache_pushLevel(self):
        self._additions = {} # dict/list attrib -> addition list
        for k in self.WrappedDicts + self.WrappedLists: self._additions[k] = []
        self._macroAccessSet = set()
        self._macroAddSet = set()
        self._filenames = set()
        self._cache_stack.append(
            MyDict(
                oldMacros = dict(self._stateStruct.macros),
                additions = self._additions,
                macroAccessSet = self._macroAccessSet,
                macroAddSet = self._macroAddSet,
                filenames = self._filenames
                ))
    def cache_popLevel(self):
        cache_data = self._cache_stack.pop()
        if len(self._cache_stack) == 0:
            del self._additions
            del self._macroAccessSet
            del self._macroAddSet
            del self._filenames
        else:
            # recover last
            last = self._cache_stack[-1]
            self._additions = last.additions
            self._macroAccessSet = last.macroAccessSet
            self._macroAddSet = last.macroAddSet
            self._filenames = last.filenames
            # merge with popped frame
            for k in self.WrappedDicts + self.WrappedLists:
                self._additions[k].extend(cache_data.additions[k])
            self._macroAddSet.update(cache_data.macroAddSet)
            for k in cache_data.macroAccessSet:
                if k not in self._macroAddSet:
                    self._macroAccessSet.add(k)
            self._filenames.update(cache_data.filenames)
        return cache_data
    preprocess = State__cached_preprocess
    def __getstate__(self):
        # many C structure objects refer to this as their parent.
        # when we pickle those objects, it should be safe to ignore to safe this.
        # we also don't really have any other option because we don't want to
        # dump this whole object.
        return None

def parse(filename, state = None):
    if state is None:
        state = cparser.State()
        state.autoSetupSystemMacros()

    wrappedState = StateWrapper(state)
    preprocessed = wrappedState.preprocess_file(filename, local=True)
    tokens = cparser.cpre2_parse(wrappedState, preprocessed)
    cparser.cpre3_parse(wrappedState, tokens)

    return state

def test():
    import better_exchook
    better_exchook.install()

    state = parse("/Library/Frameworks/SDL.framework/Headers/SDL.h")

    return state

if __name__ == '__main__':
    print(test())



import wx
import wx.lib.agw.aui as aui

class AuiToolBarPopup(wx.Frame):
    def __init__(self, parent):
        wx.Frame.__init__(self, parent, style=wx.NO_BORDER|
                          wx.FRAME_TOOL_WINDOW|wx.FRAME_NO_TASKBAR|
                          wx.FRAME_FLOAT_ON_PARENT|wx.STAY_ON_TOP)
        self._toolbar = None
        self.tb = aui.AuiToolBar(self, -1, agwStyle=aui.AUI_TB_VERTICAL|
                                 aui.AUI_TB_TEXT|aui.AUI_TB_HORZ_TEXT)
        self.Bind(wx.EVT_ACTIVATE, self.OnActivate)

    def OnTool(self, event):
        item = self.tb.FindTool(event.GetId())
        if item.GetKind() in [aui.ITEM_CHECK, aui.ITEM_RADIO]:
            state = (item.state & aui.AUI_BUTTON_STATE_CHECKED and [True] or [False])[0]
            self._toolbar.ToggleTool(item.GetId(), state)
        self.Cancel()
        event.Skip()

    def OnActivate(self, event):
        if not event.GetActive():
            wnd = wx.FindWindowAtPointer()
            if wnd:
                wnd = wnd[0]
            while wnd:
                if wnd == self:
                    return
                wnd = wnd.GetParent()
            self.Cancel()
        event.Skip()

    def Cancel(self):
        realize = False
        for item in self.tb._items:
            if item.GetKind() == aui.ITEM_CONTROL:
                origin_item = self._toolbar.FindTool(item.window.GetId())
                if origin_item.sizer_item and not origin_item.sizer_item.IsShown():
                    item.window.Show(False)
                item.window.Reparent(self._toolbar)
                realize = True
        if realize:
            # add controls back to the original toolbar, so realize it;
            # however, the client size will set to be the min size (of all
            # items); call SetClientSize() does not work since it will trigger
            # OnSize, which call Realize() again and set the client size to min
            # size again... The following "workaround" seems working
            sz = self._toolbar.GetClientSize()
            self._toolbar.Realize()
            if self._toolbar.GetAuiManager():
                self._toolbar.GetAuiManager().Update()
            self._toolbar.SetClientSize(sz)
        # refresh the toolbar; otherwise, the status of the radio buttons may
        # not be correct.
        self._toolbar.Refresh()
        self.tb.ClearTools()
        wx.CallAfter(self.Hide)

    def UpdateItems(self, wnd, items):
        items_added = 0
        self._toolbar = wnd
        self.Unbind(wx.EVT_TOOL)
        self.tb.ClearTools()

        agw_flag_wnd = wnd.GetAGWWindowStyleFlag()
        agw_flag = aui.AUI_TB_VERTICAL|aui.AUI_TB_TEXT|aui.AUI_TB_HORZ_TEXT
        if agw_flag_wnd & aui.AUI_TB_PLAIN_BACKGROUND:
            agw_flag |= aui.AUI_TB_PLAIN_BACKGROUND
        self.tb.SetAGWWindowStyleFlag(agw_flag)

        for item in items:
            tool = None
            if item.GetKind() == aui.ITEM_LABEL:
                tool = self.tb.AddLabel(item.GetId(), item.GetLabel())
            elif item.GetKind() == aui.ITEM_SEPARATOR:
                if items_added > 0:
                    tool = self.tb.AddSeparator()
            elif item.GetKind() == aui.ITEM_CONTROL:
                item.window.Reparent(self.tb)
                if not item.window.IsShown():
                    item.window.Show(True)
                tool = self.tb.AddControl(item.window, item.GetLabel())
            elif item.GetKind() == aui.ITEM_CHECK:
                tool = self.tb.AddCheckTool(item.GetId(), item.GetLabel(),
                                            item.GetBitmap(), item.GetDisabledBitmap())
                if item.state & aui.AUI_BUTTON_STATE_CHECKED:
                    tool.state |= aui.AUI_BUTTON_STATE_CHECKED
            elif item.GetKind() == aui.ITEM_RADIO:
                tool = self.tb.AddRadioTool(item.GetId(), item.GetLabel(),
                                            item.GetBitmap(), item.GetDisabledBitmap())
                if item.state & aui.AUI_BUTTON_STATE_CHECKED:
                    tool.state |= aui.AUI_BUTTON_STATE_CHECKED
            elif item.GetKind() == aui.ITEM_NORMAL:
                tool = self.tb.AddSimpleTool(item.GetId(), item.GetLabel(),
                                             item.GetBitmap())
                self.tb.SetToolDropDown(tool.GetId(), wnd.GetToolDropDown(tool.GetId()))
            if tool:
                tool.SetAlignment(wx.EXPAND)
                items_added += 1

        self.tb.Realize()
        self.SetClientSize(self.tb.GetMinSize())
        self.Bind(wx.EVT_TOOL, self.OnTool)

class AuiToolBarPopupArt(aui.AuiDefaultToolBarArt):
    def __init__(self, frame):
        aui.AuiDefaultToolBarArt.__init__(self)
        self.popup = AuiToolBarPopup(frame)

    def ShowDropDown(self, wnd, items):
        """
        Shows the drop down window menu for overflow items.

        :param `wnd`: an instance of :class:`wx.Window`;
        :param list `items`: a list of the overflow toolbar items.
        """
        # find out where to put the popup menu of window items
        pt = wx.GetMousePosition()
        pt = wnd.ScreenToClient(pt)

        # find out the screen coordinate at the bottom of the tab ctrl
        cli_rect = wnd.GetClientRect()
        pt.y = cli_rect.y + cli_rect.height
        self.popup.Position = wnd.ClientToScreen(pt)
        self.popup.UpdateItems(wnd, items)
        self.popup.Show()
        self.popup.Raise()
        return -1



import numpy as np
import wx
import wx.lib.agw.aui as aui
import wx.py as py
from . import propgrid as pg
from . import formatters as fmt
from . import enumtype
from . import propart as pa
from .propxpm import radio_xpm, tree_xpm


def BitmapFromXPM(xpm):
    xpm_b = [x.encode('utf-8') for x in xpm]
    return wx.Bitmap(xpm_b)


class PropArtCustom(pa.PropArtNative):
    def __init__(self):
        super(PropArtCustom, self).__init__()

        self.img_expand = wx.ImageList(12, 12, True, 2)
        self.img_expand.Add(BitmapFromXPM(tree_xpm))

        self.expansion_width = 12
        self.text_clr_sel = self.text_clr

    def DrawExpansion(self, dc, p):
        if p.HasChildren():
            if self.img_expand.GetImageCount() == 2:
                (w, h) = self.img_expand.GetSize(0)
                rc = p.regions['expander']
                x = rc.x + (rc.width - w) / 2
                y = rc.y + (rc.height - h) / 2 + 1
                idx = 0
                if not p.IsExpanded():
                    idx = 1
                self.img_expand.Draw(idx, dc, x, y,
                                     wx.IMAGELIST_DRAW_TRANSPARENT)
            else:
                super(PropArtCustom, self).DrawExpansion(dc, p)

    def DrawLabel(self, dc, p):
        # draw label
        if p.font_label is None:
            font = self._font_label
        else:
            font = wx.Font(p.font_label)
        if p.IsActivated():
            dc.SetFont(font.Bold())
        else:
            dc.SetFont(font)

        if not p.IsEnabled() or p.IsReadonly():
            clr = wx.SystemSettings.GetColour(wx.SYS_COLOUR_GRAYTEXT)
        else:
            clr = wx.SystemSettings.GetColour(wx.SYS_COLOUR_BTNTEXT)
        dc.SetTextForeground(clr)
        rc = p.regions['label']
        dc.SetClippingRegion(rc)
        (w, h) = dc.GetTextExtent(p.label)

        dc.DrawText(p.label, rc.x, rc.y + (rc.height - h) / 2)
        p.show_label_tips = w > rc.width
        dc.DestroyClippingRegion()

    def DrawSplitter(self, dc, p):
        # draw splitter
        rcs = p.regions['splitter']
        dc.SetPen(wx.Pen(wx.SystemSettings.GetColour(wx.SYS_COLOUR_3DSHADOW)))
        dc.DrawLine(rcs.right - 1, rcs.top, rcs.right - 1, rcs.bottom)
        dc.SetPen(wx.Pen(wx.SystemSettings.GetColour(wx.SYS_COLOUR_3DHILIGHT)))
        dc.DrawLine(rcs.right, rcs.top, rcs.right, rcs.bottom)

    def DrawBackground(self, dc, p):
        # draw background
        rc = p.GetRect()
        rcs = p.regions['splitter']
        bg = wx.SystemSettings.GetColour(wx.SYS_COLOUR_3DFACE)
        brush = wx.Brush(bg)
        dc.SetBrush(brush)
        dc.DrawRectangle(rc.x, rc.y, rcs.right, rc.height)

    def DrawValue(self, dc, p):
        # draw value
        if p.font_value is None:
            font = self._font_value
        else:
            font = wx.Font(p.font_value)
        dc.SetFont(font)

        p.show_value_tips = False
        if p.window is None:
            crbg = p.bg_clr
            crtxt = wx.BLACK
            if not p.IsEnabled() or p.IsReadonly():
                crtxt = p.text_clr_disabled
                if not crtxt:
                    crtxt = self.text_clr_disabled
                crbg = p.bg_clr_disabled
                if not crbg:
                    crbg = self.bg_clr_disabled
            elif p.activated:
                crtxt = p.text_clr_sel
                if not crtxt:
                    crtxt = self.text_clr_sel
                crbg = p.bg_clr_sel
                if not crbg:
                    crbg = self.bg_clr_sel
            else:
                crtxt = p.text_clr
                if not crtxt:
                    crtxt = self.text_clr
                crbg = p.bg_clr
                if not crbg:
                    crbg = self.bg_clr
            c = wx.Colour(crbg)
            dc.SetPen(wx.Pen(wx.Colour(c), 1, wx.PENSTYLE_SOLID))
            rc = p.regions['value']
            rcbg = wx.Rect(*rc)

            if p.activated:
                dc.SetBrush(wx.Brush(wx.Colour(c.red, c.green, c.blue, 128)))
                rcbg.Deflate(0, 1)
            else:
                dc.SetBrush(wx.Brush(wx.Colour(c.red, c.green, c.blue, 255)))

            dc.DrawRectangle(rcbg)

            dc.SetPen(wx.Pen(crtxt, 1, wx.PENSTYLE_TRANSPARENT))
            dc.SetTextForeground(crtxt)

            value = p.GetValueAsString()
            (w, h) = dc.GetTextExtent(value)
            dc.SetClippingRegion(rc)
            dc.DrawText(value, rc.x + 5, rc.top + (rc.height - h) / 2)
            p.show_value_tips = rc.width < w
            dc.DestroyClippingRegion()

    def DrawBorder(self, dc, p):
        rc = p.GetRect()
        rcs = p.regions['splitter']
        # value bottom border
        crbg = p.bg_clr
        if p.IsEnabled() and not p.IsReadonly():
            crbg = p.bg_clr_disabled
            if not crbg:
                crbg = self.bg_clr_disabled
        else:
            crbg = p.bg_clr
            if not crbg:
                crbg = self.bg_clr

        dc.SetPen(wx.Pen(wx.Colour(crbg)))
        dc.DrawLine(rc.left, rc.bottom, rc.right, rc.bottom)

        # title top & bottom border
        dc.SetPen(wx.Pen(wx.SystemSettings.GetColour(wx.SYS_COLOUR_3DSHADOW)))
        dc.DrawLine(rc.left, rc.bottom, rcs.right, rc.bottom)
        dc.SetPen(wx.Pen(wx.SystemSettings.GetColour(wx.SYS_COLOUR_3DHILIGHT)))
        dc.DrawLine(rc.left, rc.top, rcs.right, rc.top)


class MainFrame(wx.Frame):
    ID_ART_NATIVE = wx.NewId()
    ID_ART_DEFAULT = wx.NewId()

    def __init__(self):
        wx.Frame.__init__(self, None, -1, 'PropGrid Demo', size=(800, 600))
        self._mgr = aui.AuiManager()

        # tell AuiManager to manage this frame
        self._mgr.SetManagedWindow(self)
        self._mgr.SetAGWFlags(self._mgr.GetAGWFlags()
                              | aui.AUI_MGR_ALLOW_ACTIVE_PANE
                              | aui.AUI_MGR_SMOOTH_DOCKING
                              | aui.AUI_MGR_USE_NATIVE_MINIFRAMES
                              | aui.AUI_MGR_LIVE_RESIZE)
        menubar = wx.MenuBar()
        viewMenu = wx.Menu()
        item = wx.MenuItem(viewMenu,
                           self.ID_ART_NATIVE,
                           text="Native Art",
                           kind=wx.ITEM_CHECK)
        viewMenu.Append(item)
        item = wx.MenuItem(viewMenu,
                           self.ID_ART_DEFAULT,
                           text="Customized Art",
                           kind=wx.ITEM_CHECK)
        viewMenu.Append(item)
        menubar.Append(viewMenu, '&Options')
        self.SetMenuBar(menubar)

        self.Bind(wx.EVT_TOOL, self.OnProcessTool)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateCmdUI)

        self.propgrid = pg.PropGrid(self)
        g = self.propgrid
        # general
        p = g.InsertSeparator('general', 'general')

        p = g.InsertProperty('string', 'string', 'hello world!')
        p.SetIndent(1)

        p = g.InsertProperty('disable', 'disable', 'hello world!')
        p.SetIndent(1)
        p.Enable(False)

        p = g.InsertProperty('italic', 'italic', 'hello world!')
        p.SetIndent(1)
        p.SetValueFont(wx.Font(wx.NORMAL_FONT).Italic())

        p = g.InsertProperty('bold', 'bold', 'hello world!')
        p.SetIndent(1)
        p.SetValueFont(wx.Font(wx.NORMAL_FONT).Bold())

        p = g.InsertProperty('integer', 'integer', 42)
        p.SetIndent(1)
        p.SetFormatter(fmt.IntFormatter())

        p = g.InsertProperty('hex', 'hex', 42)
        p.SetIndent(1)
        p.SetFormatter(fmt.HexFormatter())

        p = g.InsertProperty('bin', 'bin', 42)
        p.SetIndent(1)
        p.SetFormatter(fmt.BinFormatter())

        p = g.InsertProperty('float', 'float', 42.0)
        p.SetIndent(1)
        p.SetFormatter(fmt.FloatFormatter())

        p = g.InsertProperty('choice', 'choice', 2)
        p.SetFormatter(fmt.ChoiceFormatter([2, 4, 8, 16, 32, 64, 128, 256]))
        p.SetIndent(1)

        p = g.InsertProperty('date', 'date', wx.DateTime.Today())
        p.SetIndent(1)
        p.SetFormatter(fmt.DateFormatter())

        p = g.InsertProperty('time', 'time', wx.DateTime.Now())
        p.SetIndent(1)
        p.SetFormatter(fmt.TimeFormatter())

        p = g.InsertProperty('datetime', 'datetime', wx.DateTime.Now())
        p.SetIndent(1)
        p.SetFormatter(fmt.DateTimeFormatter())
        p.SetControlStyle('none')

        # control
        p = g.InsertSeparator('type', 'type')

        p = g.InsertProperty('editbox', 'editbox', 'string')
        p.SetIndent(1)

        p = g.InsertProperty('choice', 'choice', 1)
        choices = enumtype.EnumType(Monday=1,
                                    Tuesday=2,
                                    Wednesday=3,
                                    Thursday=4,
                                    Friday=5,
                                    Saturday=6,
                                    Sunday=7)
        p.SetFormatter(fmt.EnumFormatter(choices))
        p.SetIndent(1)

        p = g.InsertProperty('dir_dialog', 'folder', '/home')
        #p.SetControlStyle('dir_dialog')
        p.SetFormatter(fmt.PathFormatter(False, 'folder'))
        p.SetIndent(1)

        p = g.InsertProperty('file_dialog', 'file', '/home/temp.txt')
        #p.SetControlStyle('file_dialog')
        p.SetFormatter(fmt.PathFormatter(False, 'file'))
        p.SetIndent(1)

        p = g.InsertProperty('slider', 'slider', 50)
        p.SetControlStyle('slider')
        p.SetFormatter(fmt.IntFormatter(1, 101))
        p.SetIndent(1)

        p = g.InsertProperty('spin', 'spin', 50)
        p.SetControlStyle('spin')
        p.SetFormatter(fmt.IntFormatter(1, 100))
        p.SetIndent(1)

        p = g.InsertProperty('checkbox', 'checkbox', 0)
        p.SetControlStyle('checkbox')
        p.SetFormatter(fmt.BoolFormatter())
        p.SetIndent(1)

        p = g.InsertProperty('radiobox', 'radiobox', 1)
        #p.SetFormatter(fmt.EnumFormatter(choices))
        p.SetFormatter(
            fmt.ChoiceFormatter({
                1: '1',
                0: '0',
                'Z': 'Z',
                'X': 'X'
            }))
        p.SetControlStyle('radiobox')
        p.SetIndent(1)

        p = g.InsertProperty('date', 'date', wx.DateTime.Today())
        p.SetIndent(1)
        p.SetFormatter(fmt.DateFormatter())

        p = g.InsertProperty('time', 'time', wx.DateTime.Now())
        p.SetIndent(1)
        p.SetFormatter(fmt.TimeFormatter())

        p = g.InsertProperty('font', 'font', wx.NORMAL_FONT)
        p.SetIndent(1)
        p.SetFormatter(fmt.FontFormatter())

        # color
        self.clr_map = np.array(
            [[170., 110., 40., 255.], [0., 0., 128., 255.],
             [255., 225., 25., 255.], [128., 128., 128., 255.],
             [128., 0., 0., 255.], [0., 128., 128., 255.], [0., 0., 0., 255.],
             [210., 245., 60., 255.], [250., 190., 190., 255.],
             [145., 30., 180., 255.], [128., 128., 0., 255.],
             [240., 50., 230., 255.], [230., 25., 75., 255.],
             [255., 255., 255., 255.], [230., 190., 255., 255.],
             [255., 215., 180., 255.], [60., 180., 75., 255.],
             [255., 250., 200., 255.], [245., 130., 48., 255.],
             [0., 130., 200., 255.], [70., 240., 240., 255.],
             [170., 255., 195., 255.]]) / 255
        chex = self.rgb2hex(self.clr_map[:, :-1])
        p = self.propgrid.InsertSeparator("color", "color")
        for i, c in enumerate(chex):
            p = self.propgrid.InsertProperty("clr-%d" % i, '%d' % i,
                                             wx.Colour(c))
            p.SetBgColor(c, c, c)
            p.SetFormatter(fmt.ColorFormatter())
            t = wx.Colour(c)
            t.SetRGB(t.GetRGB() ^ 0xFFFFFF)
            t = t.GetAsString(wx.C2S_HTML_SYNTAX)
            p.SetTextColor(t, t, t)
            p.SetIndent(1)

        pane_grid = aui.AuiPaneInfo().Name("propgrid").Caption("PropGrid").\
                    CenterPane()
        self._mgr.AddPane(self.propgrid, pane_grid)
        ns = {}
        ns['wx'] = wx
        ns['app'] = wx.GetApp()
        ns['frame'] = self
        self.shell = py.shell.Shell(self, -1, locals=ns)

        pane_shell = aui.AuiPaneInfo().Name("Shell").Caption("Shell").Bottom()\
                          .BestSize((300, 300)).DestroyOnClose(False).Snappable()\
                          .Dockable().MinimizeButton(True).MaximizeButton(True)\
                          .Row(-1).Bottom().Position(99)

        self._mgr.AddPane(self.shell, pane_shell)
        self._mgr.ShowPane(g, True)
        self._mgr.MinimizePane(pane_shell)
        self._mgr.Update()

        self.Bind(pg.EVT_PROP_CHANGED, self.OnPropChanged, id=wx.ID_ANY)

        self.timer = wx.Timer(self)
        self.Bind(wx.EVT_TIMER, self.OnTimer, self.timer)
        self.timer.Start(100)

    def rgb2hex(self, clr):
        clr = np.sum(clr * 255 * [2**16, 2**8, 1], 1).astype(np.int32)
        return ["#{:06x}".format(c) for c in clr]

    def OnPropChanged(self, evt):
        p = evt.GetProperty()
        if p.GetName().startswith('clr-'):
            idx = int(p.GetLabel())
            if idx >= 0 and idx < len(self.clr_map):
                t = wx.Colour(p.GetValue())
                self.clr_map[idx] = [
                    t.Red() / 255.,
                    t.Green() / 255.,
                    t.Blue() / 255., 1.
                ]
                c = t.GetAsString(wx.C2S_HTML_SYNTAX)
                p.SetBgColor(c, c, c)
                t.SetRGB(t.GetRGB() ^ 0xFFFFFF)
                t = t.GetAsString(wx.C2S_HTML_SYNTAX)
                p.SetTextColor(t, t, t)
        if 'font' in p.GetName():
            p.SetValueFont(p.GetValue())

    def OnTimer(self, event):
        p = self.propgrid.GetProperty('datetime')
        if p:
            p.SetValue(wx.DateTime.Now())

    def OnUpdateCmdUI(self, event):
        eid = event.GetId()
        if eid == self.ID_ART_NATIVE:
            event.Check(
                type(self.propgrid.GetArtProvider()) == pa.PropArtNative)
        elif eid == self.ID_ART_DEFAULT:
            event.Check(type(self.propgrid.GetArtProvider()) == PropArtCustom)
        else:
            event.Skip()

    def OnProcessTool(self, event):
        eid = event.GetId()
        if eid == self.ID_ART_NATIVE:
            self.propgrid.SetArtProvider(pa.PropArtNative())
            self.propgrid.SetBackgroundColour(wx.NullColour)
        elif eid == self.ID_ART_DEFAULT:
            self.propgrid.SetArtProvider(PropArtCustom())
            self.propgrid.SetBackgroundColour(wx.WHITE)
        else:
            event.Skip()


class RunApp(wx.App):
    def __init__(self):
        wx.App.__init__(self, redirect=False)

    def OnInit(self):
        frame = MainFrame()
        frame.Show(True)
        self.SetTopWindow(frame)
        self.frame = frame
        return True


def main():
    app = RunApp()
    app.MainLoop()


if __name__ == '__main__':
    main()



import os
import inspect
import keyword
import pprint
import six
import wx
import wx.stc as stc
import wx.py.dispatcher as dp
import wx.lib.agw.aui as aui
from ..auibarpopup import AuiToolBarPopupArt
from .bsmxpm import open_xpm, save_xpm, saveas_xpm, find_xpm, indent_xpm, \
                    dedent_xpm, run_xpm, execute_xpm, check_xpm, debug_xpm, \
                    folder_xpm, vert_xpm, horz_xpm, reload_xpm
from .pymgr_helpers import Gcm
from .. import to_byte


class BreakpointSettingsDlg(wx.Dialog):
    def __init__(self, parent, condition='', hitcount='', curhitcount=0):
        wx.Dialog.__init__(self,
                           parent,
                           title="Breakpoint Condition",
                           size=wx.Size(431, 290),
                           style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER)

        self.SetSizeHintsSz(wx.DefaultSize, wx.DefaultSize)
        szAll = wx.BoxSizer(wx.VERTICAL)
        label = ('When the breakkpoint location is reached, the expression is '
                 'evaluated and the breakpoint is hit only if the expression '
                 'is true.')
        self.stInfo = wx.StaticText(self, label=label)
        self.stInfo.Wrap(-1)
        szAll.Add(self.stInfo, 1, wx.ALL, 15)
        szCnd = wx.BoxSizer(wx.HORIZONTAL)

        szCnd.AddSpacer((20, 0), 0, wx.EXPAND, 5)

        szCond = wx.BoxSizer(wx.VERTICAL)

        self.cbCond = wx.CheckBox(self, label="Is true")
        szCond.Add(self.cbCond, 0, wx.ALL | wx.EXPAND, 5)

        self.tcCond = wx.TextCtrl(self, wx.ID_ANY)
        szCond.Add(self.tcCond, 0, wx.ALL | wx.EXPAND, 5)

        label = "Hit count (hit count: #; for example, #>10"
        self.cbHitCount = wx.CheckBox(self, label=label)
        szCond.Add(self.cbHitCount, 0, wx.ALL, 5)

        self.tcHitCount = wx.TextCtrl(self, wx.ID_ANY)
        szCond.Add(self.tcHitCount, 0, wx.ALL | wx.EXPAND, 5)
        label = "Current hit count: %d" % curhitcount
        self.stHtCount = wx.StaticText(self, label=label)
        szCond.Add(self.stHtCount, 0, wx.ALL | wx.EXPAND, 5)

        szCnd.Add(szCond, 1, wx.EXPAND, 5)

        szAll.Add(szCnd, 1, wx.EXPAND, 5)

        self.stLine = wx.StaticLine(self, style=wx.LI_HORIZONTAL)
        szAll.Add(self.stLine, 0, wx.EXPAND | wx.ALL, 5)

        szConfirm = wx.BoxSizer(wx.HORIZONTAL)

        self.btnOK = wx.Button(self, wx.ID_OK, "OK")
        szConfirm.Add(self.btnOK, 0, wx.ALL, 5)

        self.btnCancel = wx.Button(self, wx.ID_CANCEL, "Cancel")
        szConfirm.Add(self.btnCancel, 0, wx.ALL, 5)

        szAll.Add(szConfirm, 0, wx.ALIGN_RIGHT, 5)

        self.SetSizer(szAll)
        self.Layout()

        # initialize the controls
        self.condition = condition
        self.hitcount = hitcount
        self.SetSizer(szAll)
        self.Layout()
        if self.condition == '':
            self.cbCond.SetValue(False)
            self.tcCond.Disable()
        else:
            self.cbCond.SetValue(True)
        self.tcCond.SetValue(self.condition)
        if self.hitcount == '':
            self.cbHitCount.SetValue(False)
            self.tcHitCount.Disable()
        else:
            self.cbHitCount.SetValue(True)
        self.tcHitCount.SetValue(self.hitcount)
        # Connect Events
        self.cbCond.Bind(wx.EVT_CHECKBOX, self.OnRadioButton)
        self.cbHitCount.Bind(wx.EVT_CHECKBOX, self.OnRadioButton)
        self.btnOK.Bind(wx.EVT_BUTTON, self.OnBtnOK)

    def OnRadioButton(self, event):
        self.tcCond.Enable(self.cbCond.GetValue())
        self.tcHitCount.Enable(self.cbHitCount.GetValue())
        event.Skip()

    def OnBtnOK(self, event):
        # set condition to empty string to indicate the breakpoint will be
        # trigged when the value is changed
        if self.cbCond.GetValue():
            self.condition = self.tcCond.GetValue()
        else:
            self.condition = ''
        if self.cbHitCount.GetValue():
            self.hitcount = self.tcHitCount.GetValue()
        else:
            self.hitcount = ""
        event.Skip()

    def GetCondition(self):
        return (self.condition, self.hitcount)


NUM_MARGIN = 0
MARK_MARGIN = 1
FOLD_MARGIN = 2


class PyEditor(wx.py.editwindow.EditWindow):
    ID_COMMENT = wx.NewId()
    ID_UNCOMMENT = wx.NewId()
    ID_EDIT_BREAKPOINT = wx.NewId()
    ID_DELETE_BREAKPOINT = wx.NewId()
    ID_CLEAR_BREAKPOINT = wx.NewId()

    def __init__(self, parent, style=wx.CLIP_CHILDREN | wx.BORDER_NONE):
        wx.py.editwindow.EditWindow.__init__(self, parent, style=style)
        self.SetUpEditor()
        # disable the auto-insert the call tip
        self.callTipInsert = False
        self.filename = ""
        self.autoCompleteKeys = [ord('.')]
        rsp = dp.send('shell.auto_complete_keys')
        if rsp:
            self.autoCompleteKeys = rsp[0][1]
        self.breakpointlist = {}
        self.highlightStr = ""
        self.SetMouseDwellTime(500)

        # Assign handlers for keyboard events.
        self.Bind(wx.EVT_CHAR, self.OnChar)
        self.Bind(wx.EVT_KEY_DOWN, self.OnKeyDown)
        self.Bind(wx.EVT_KILL_FOCUS, self.OnKillFocus)
        self.Bind(wx.EVT_LEFT_UP, self.OnLeftUp)
        self.Bind(stc.EVT_STC_MARGINCLICK, self.OnMarginClick)
        self.Bind(stc.EVT_STC_DOUBLECLICK, self.OnDoubleClick)
        self.Bind(stc.EVT_STC_DWELLSTART, self.OnMouseDwellStart)
        self.Bind(stc.EVT_STC_DWELLEND, self.OnMouseDwellEnd)
        # Assign handler for the context menu
        self.Bind(wx.EVT_CONTEXT_MENU, self.OnContextMenu)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateCommandUI)
        self.Bind(wx.EVT_MENU, self.OnProcessEvent)

    def ClearBreakpoint(self):
        """clear all the breakpoint"""
        for key in list(self.breakpointlist):
            ids = self.breakpointlist[key]['id']
            dp.send('debugger.clear_breakpoint', id=ids)

    def SaveFile(self, filename):
        """save file"""
        if super(PyEditor, self).SaveFile(filename):
            # remember the filename
            fname = os.path.abspath(filename)
            fname = os.path.normcase(fname)
            self.filename = fname
            return True
        return False

    def LoadFile(self, filename):
        """load file into editor"""
        self.ClearBreakpoint()
        if super(PyEditor, self).LoadFile(filename):
            # remember the filename
            fname = os.path.abspath(filename)
            fname = os.path.normcase(fname)
            self.filename = fname
            return True
        return False

    def OnKillFocus(self, event):
        """lose focus"""
        # close the autocomplete and calltip windows
        if self.CallTipActive():
            self.CallTipCancel()
        # crash on mac
        #if self.AutoCompActive():
        #    self.AutoCompCancel()
        event.Skip()

    def OnChar(self, event):
        """
        Keypress event handler.

        Only receives an event if OnKeyDown calls event.Skip() for the
        corresponding event.
        """
        key = event.GetKeyCode()
        currpos = self.GetCurrentPos()
        line = self.GetCurrentLine()
        stoppos = self.PositionFromLine(line)
        # Return (Enter) needs to be ignored in this handler.
        if key in [wx.WXK_RETURN, wx.WXK_NUMPAD_ENTER]:
            pass
        elif key in self.autoCompleteKeys:
            # Usually the dot (period) key activates auto completion.
            # Get the command between the prompt and the cursor.  Add
            # the autocomplete character to the end of the command.
            if self.AutoCompActive():
                self.AutoCompCancel()
            command = self.GetTextRange(stoppos, currpos) + chr(key)
            self.AddText(chr(key))
            if self.autoComplete:
                self.autoCompleteShow(command)
        elif key == ord('('):
            # The left paren activates a call tip and cancels an
            # active auto completion.
            if self.AutoCompActive():
                self.AutoCompCancel()
            # Get the command between the prompt and the cursor.  Add
            # the '(' to the end of the command.
            self.ReplaceSelection("""""")
            command = self.GetTextRange(stoppos, currpos) + '('
            self.AddText('(')
            self.autoCallTipShow(command,
                                 self.GetCurrentPos() == self.GetTextLength())
        else:
            # Allow the normal event handling to take place.
            event.Skip()

    def OnKeyDown(self, event):
        """key down"""
        key = event.GetKeyCode()
        control = event.ControlDown()
        # shift=event.ShiftDown()
        alt = event.AltDown()
        if key == wx.WXK_RETURN and not control and not alt \
            and not self.AutoCompActive():
            # auto-indentation
            if self.CallTipActive():
                self.CallTipCancel()
            line = self.GetCurrentLine()
            txt = self.GetLine(line)
            pos = self.GetCurrentPos()
            linePos = self.PositionFromLine(line)
            self.CmdKeyExecute(stc.STC_CMD_NEWLINE)
            indent = self.GetLineIndentation(line)
            tabWidth = max(1, self.GetTabWidth())
            if self.GetUseTabs():
                indentation = '\t'
            else:
                indentation = ' ' * tabWidth
            padding = indentation * (int(indent / tabWidth))
            newpos = self.GetCurrentPos()
            # smart indentation
            stripped = txt[:pos - linePos].split('#')[0].strip()
            firstWord = stripped.split(' ')[0]
            if stripped and self.needsIndent(firstWord, lastChar=stripped[-1]):
                padding += indentation
            elif self.needsDedent(firstWord):
                padding = padding[:-tabWidth]
            self.InsertText(newpos, padding)
            newpos += len(padding)
            self.SetCurrentPos(newpos)
            self.SetSelection(newpos, newpos)
        else:
            event.Skip()

    def OnMarginClick(self, evt):
        """left mouse button click on margin"""
        margin = evt.GetMargin()
        ctrldown = evt.GetControl()
        # set/edit/delete a breakpoint
        if margin in [NUM_MARGIN, MARK_MARGIN]:
            lineClicked = self.LineFromPosition(evt.GetPosition())
            txt = self.GetLine(lineClicked)
            txt = txt.strip()
            if not txt or txt[0] == '#':
                return
            # check if a breakpoint marker is at this line
            bpset = self.MarkerGet(lineClicked) & 1
            bpdata = None
            resp = dp.send('debugger.get_breakpoint',
                           filename=self.filename,
                           lineno=lineClicked + 1)
            if resp:
                bpdata = resp[0][1]
            if not bpdata:
                # No breakpoint at this line, add one
                # bpdata =  {id, filename, lineno, condition, ignore_count, trigger_count}
                bp = {'filename': self.filename, 'lineno': lineClicked + 1}
                dp.send('debugger.set_breakpoint', bpdata=bp)
            else:
                if ctrldown:
                    condition = """"""
                    if bpdata['condition']:
                        condition = bpdata['condition']
                    dlg = wx.TextEntryDialog(self,
                                             caption='Breakpoint Condition:',
                                             message='Condition',
                                             defaultValue="""""",
                                             style=wx.OK)
                    if dlg.ShowModal() == wx.ID_OK:
                        dp.send('debugger.edit_breakpoint',
                                id=bpdata['id'],
                                condition=dlg.GetValue())
                else:
                    dp.send('debugger.clear_breakpoint', id=bpdata['id'])
        # fold and unfold as needed
        if evt.GetMargin() == FOLD_MARGIN:
            if evt.GetShift() and evt.GetControl():
                self.FoldAll()
            else:
                lineClicked = self.LineFromPosition(evt.GetPosition())
                level = self.GetFoldLevel(lineClicked)
                if level & stc.STC_FOLDLEVELHEADERFLAG:
                    if evt.GetShift():
                        # expand node and all subnodes
                        self.SetFoldExpanded(lineClicked, True)
                        self.Expand(lineClicked, True, True, 100, level)
                    elif evt.GetControl():
                        if self.GetFoldExpanded(lineClicked):
                            # collapse all subnodes
                            self.SetFoldExpanded(lineClicked, False)
                            self.Expand(lineClicked, False, True, 0, level)
                        else:
                            # expand all subnodes
                            self.SetFoldExpanded(lineClicked, True)
                            self.Expand(lineClicked, True, True, 100, level)
                    else:
                        self.ToggleFold(lineClicked)

    def OnMouseDwellStart(self, event):
        resp = dp.send(signal='debugger.get_status')
        if not resp or not resp[0][1]:
            return

        pos = event.GetPosition()
        #line = self.LineFromPosition(pos) # 0, 1, 2
        if pos == -1:
            return
        c = self.GetWordChars()
        self.SetWordChars(c + '.')
        WordStart = self.WordStartPosition(pos, True)
        WordEnd = self.WordEndPosition(pos, True)
        text = self.GetTextRange(WordStart, WordEnd)
        self.SetWordChars(c)
        try:
            status = resp[0][1]
            frames = status['frames']
            level = status['active_scope']
            frame = frames[level]
            f_globals = frame.f_globals
            f_locals = frame.f_locals

            tip = pprint.pformat(eval(text, f_globals, f_locals))
            self.CallTipShow(pos, "%s = %s" % (text, tip))
        except:
            #traceback.print_exc(file=sys.stdout)
            pass

    def OnMouseDwellEnd(self, event):
        if self.CallTipActive():
            self.CallTipCancel()

    def FoldAll(self):
        """open all margin folders"""
        line_count = self.GetLineCount()
        expanding = True
        # find out if we are folding or unfolding
        for line_num in six.moves.range(line_count):
            if self.GetFoldLevel(line_num) & wx.stc.STC_FOLDLEVELHEADERFLAG:
                expanding = not self.GetFoldExpanded(line_num)
                break
        line_number = 0

        while line_number < line_count:
            level = self.GetFoldLevel(line_number)
            if level & stc.STC_FOLDLEVELHEADERFLAG and \
               (level & stc.STC_FOLDLEVELNUMBERMASK) == stc.STC_FOLDLEVELBASE:

                if expanding:
                    self.SetFoldExpanded(line_number, True)
                    line_number = self.Expand(line_number, True)
                    line_number = line_number - 1
                else:
                    lastChild = self.GetLastChild(line_number, -1)
                    self.SetFoldExpanded(line_number, False)

                    if lastChild > line_number:
                        self.HideLines(line_number + 1, lastChild)

            line_number = line_number + 1

    def Expand(self, line, do_expand, force=False, vis_levels=0, level=-1):
        """open the margin folder"""
        last_child = self.GetLastChild(line, level)
        line = line + 1

        while line <= last_child:
            if force:
                if vis_levels > 0:
                    self.ShowLines(line, line)
                else:
                    self.HideLines(line, line)
            else:
                if do_expand:
                    self.ShowLines(line, line)

            if level == -1:
                level = self.GetFoldLevel(line)

            if level & wx.stc.STC_FOLDLEVELHEADERFLAG:
                if force:
                    self.SetFoldExpanded(line, vis_levels > 1)
                    line = self.Expand(line, do_expand, force, vis_levels - 1)
                else:
                    if do_expand:
                        if self.GetFoldExpanded(line):
                            self.SetFoldExpanded(line, True)
                    line = self.Expand(line, do_expand, force, vis_levels - 1)
            else:
                line = line + 1
        return line

    def OnDoubleClick(self, event):
        """left mouse button double click"""
        # highlight all the instances of the selected text
        sel = self.GetSelectedText()
        if sel != "":
            self.highlightText(sel)

        event.Skip()

    def OnLeftUp(self, event):
        """left mouse button released"""
        # remove the highlighting when click somewhere else
        sel = self.GetSelectedText()
        if self.highlightStr != "" and sel != self.highlightStr:
            self.highlightText(self.highlightStr, False)

        event.Skip()

    def highlightText(self, strWord, highlight=True):
        """highlight the text"""
        current = 0
        position = -1
        flag = stc.STC_FIND_WHOLEWORD | stc.STC_FIND_MATCHCASE
        if not highlight:
            self.IndicatorClearRange(0, self.GetLength())
            self.highlightStr = ""
            return

        self.highlightStr = strWord
        self.SetIndicatorCurrent(0)
        while True:
            position = self.FindText(current, len(self.GetText()), strWord,
                                     flag)
            if isinstance(position, tuple):
                position = position[0] # wx ver 4.1.0 returns (start, end)
            current = position + len(strWord)
            if position == -1:
                break
            self.IndicatorFillRange(position, len(strWord))

    def needsIndent(self, firstWord, lastChar):
        '''Tests if a line needs extra indenting, i.e., if, while, def, etc '''
        # remove trailing ":" on token
        if firstWord and firstWord[-1] == ':':
            firstWord = firstWord[:-1]
        # control flow keywords
        keys = [
            'for', 'if', 'else', 'def', 'class', 'elif', 'try', 'except',
            'finally', 'while', 'with'
        ]
        return firstWord in keys and lastChar == ':'

    def needsDedent(self, firstWord):
        '''Tests if a line needs extra dedenting, i.e., break, return, etc '''
        # control flow keywords
        return firstWord in ['break', 'return', 'continue', 'yield', 'raise']

    def autoCompleteShow(self, command, offset=0):
        """Display auto-completion popup list."""
        self.AutoCompSetAutoHide(self.autoCompleteAutoHide)
        self.AutoCompSetIgnoreCase(self.autoCompleteCaseInsensitive)

        options = []
        # retrieve the auto complete list from shell
        response = dp.send('shell.auto_complete_list', command=command)
        if response:
            options = response[0][1]
        if options:
            self.AutoCompShow(offset, ' '.join(options))

    def autoCallTipShow(self, command, insertcalltip=True, forceCallTip=False):
        """Display argument spec and docstring in a popup window."""
        if self.CallTipActive():
            self.CallTipCancel()
        (argspec, tip) = (None, None)
        # retrieve the all tip from shell
        response = dp.send('shell.auto_call_tip', command=command)
        if response:
            # name, argspec, tip
            (_, argspec, tip) = response[0][1]
        if tip:
            dp.send('Shell.calltip', sender=self, calltip=tip)
        if not self.autoCallTip and not forceCallTip:
            return
        startpos = self.GetCurrentPos()
        if argspec and insertcalltip and self.callTipInsert:
            self.AddText(argspec + ')')
            endpos = self.GetCurrentPos()
            self.SetSelection(startpos, endpos)
        if argspec:
            tippos = startpos
            fallback = startpos - self.GetColumn(startpos)
            # In case there isn't enough room, only go back to the
            # fallback.
            tippos = max(tippos, fallback)
            self.CallTipShow(tippos, argspec)

    # Some methods to make it compatible with how the wxTextCtrl is used
    def SetValue(self, value):
        # if wx.USE_UNICODE:
        #    value = value.decode('iso8859_1')
        val = self.GetReadOnly()
        self.SetReadOnly(False)
        self.SetText(value)
        self.EmptyUndoBuffer()
        self.SetSavePoint()
        self.SetReadOnly(val)

    def SelectLine(self, line):
        """select the line"""
        start = self.PositionFromLine(line)
        end = self.GetLineEndPosition(line)
        self.SetSelection(start, end)

    def UpdateStatusText(self):
        """update the info on statusbar"""
        caretPos = self.GetCurrentPos()
        col = self.GetColumn(caretPos) + 1
        line = self.LineFromPosition(caretPos) + 1
        dp.send('frame.show_status_text',
                text='%d, %d' % (line, col),
                index=1,
                width=100)

    def OnUpdateUI(self, event):
        super(PyEditor, self).OnUpdateUI(event)
        wx.CallAfter(self.UpdateStatusText)

    def SetUpEditor(self):
        """
        This method carries out the work of setting up the demo editor.
        It's separate so as not to clutter up the init code.
        """
        # key binding
        self.CmdKeyAssign(ord('R'), stc.STC_SCMOD_CTRL, stc.STC_CMD_REDO)
        self.SetLexer(stc.STC_LEX_PYTHON)
        keywords = keyword.kwlist
        keywords.extend(['None', 'as', 'True', 'False'])
        self.SetKeyWords(0, ' '.join(keywords))
        # Enable folding
        self.SetProperty('fold', '1')
        # Highlight tab/space mixing (shouldn't be any)
        self.SetProperty('tab.timmy.whinge.level', '1')
        # Set left and right margins
        self.SetMargins(2, 2)
        # Set up the numbers in the margin for margin #1
        self.SetMarginType(NUM_MARGIN, stc.STC_MARGIN_NUMBER)
        # Reasonable value for, say, 4-5 digits using a mono font (40 pix)
        self.SetMarginWidth(0, 50)
        # Indentation and tab stuff
        self.SetIndent(4)
        self.SetIndentationGuides(True)
        # Backspace unindents rather than delete 1 space
        self.SetBackSpaceUnIndents(True)
        self.SetTabIndents(True)

        # Use spaces rather than tabs, or TabTimmy will complain!
        self.SetUseTabs(False)
        # Don't view white space
        self.SetViewWhiteSpace(False)
        # EOL: Since we are loading/saving ourselves, and the
        # strings will always have \n's in them, set the STC to
        # edit them that way.
        self.SetEOLMode(stc.STC_EOL_LF)
        self.SetViewEOL(False)
        # No right-edge mode indicator
        self.SetEdgeMode(stc.STC_EDGE_NONE)
        # Margin #1 - breakpoint symbols
        self.SetMarginType(MARK_MARGIN, stc.STC_MARGIN_SYMBOL)
        # do not show fold symbols
        self.SetMarginMask(MARK_MARGIN, ~stc.STC_MASK_FOLDERS)
        self.SetMarginSensitive(MARK_MARGIN, True)
        self.SetMarginWidth(MARK_MARGIN, 12)
        # break point
        self.MarkerDefine(0, stc.STC_MARK_CIRCLE, '#8E8E93', '#FF3B30')
        # paused at marker
        self.MarkerDefine(1, stc.STC_MARK_SHORTARROW, '#8E8E93', '#34C759')
        self.MarkerDefine(2, stc.STC_MARK_SHORTARROW, '#8E8E93', 'WHITE')

        # Setup a margin to hold fold markers
        self.SetMarginType(FOLD_MARGIN, stc.STC_MARGIN_SYMBOL)
        self.SetMarginMask(FOLD_MARGIN, stc.STC_MASK_FOLDERS)
        self.SetMarginSensitive(FOLD_MARGIN, True)
        self.SetMarginWidth(FOLD_MARGIN, 12)
        # and now set up the fold markers
        self.MarkerDefine(stc.STC_MARKNUM_FOLDEREND,
                          stc.STC_MARK_BOXPLUSCONNECTED, 'white', '#8E8E93')
        self.MarkerDefine(stc.STC_MARKNUM_FOLDEROPENMID,
                          stc.STC_MARK_BOXMINUSCONNECTED, 'white', '#8E8E93')
        self.MarkerDefine(stc.STC_MARKNUM_FOLDERMIDTAIL, stc.STC_MARK_TCORNER,
                          'white', '#8E8E93')
        self.MarkerDefine(stc.STC_MARKNUM_FOLDERTAIL, stc.STC_MARK_LCORNER,
                          'white', '#8E8E93')
        self.MarkerDefine(stc.STC_MARKNUM_FOLDERSUB, stc.STC_MARK_VLINE,
                          'white', '#8E8E93')
        self.MarkerDefine(stc.STC_MARKNUM_FOLDER, stc.STC_MARK_BOXPLUS,
                          'white', '#8E8E93')
        self.MarkerDefine(stc.STC_MARKNUM_FOLDEROPEN, stc.STC_MARK_BOXMINUS,
                          'white', '#8E8E93')
        # Global default style
        if wx.Platform == '__WXMSW__':
            self.StyleSetSpec(stc.STC_STYLE_DEFAULT,
                              'fore:#1D1D1D,back:#FFFFFF,face:Courier New')
        elif wx.Platform == '__WXMAC__':
            # TODO: if this looks fine on Linux too, remove the Mac-specific case
            # and use this whenever OS != MSW.
            self.StyleSetSpec(stc.STC_STYLE_DEFAULT,
                              'fore:#1D1D1D,back:#FFFFFF,face:Monaco')
        else:
            #defsize = \
            #    wx.SystemSettings.GetFont(wx.SYS_ANSI_FIXED_FONT).GetPointSize()
            defsize = 14
            self.StyleSetSpec(
                stc.STC_STYLE_DEFAULT,
                'fore:#000000,back:#FFFFFF,face:Courier,size:%d' % defsize)
        # Clear styles and revert to default.
        self.StyleClearAll()
        # Following style specs only indicate differences from default.
        # The rest remains unchanged.
        # Line numbers in margin
        self.StyleSetSpec(stc.STC_STYLE_LINENUMBER,
                          'fore:#3C3C43,back:#F2F2F7')
        # Highlighted brace
        self.StyleSetSpec(stc.STC_STYLE_BRACELIGHT,
                          'fore:#00009D,back:#FFFF00')
        # Unmatched brace
        self.StyleSetSpec(stc.STC_STYLE_BRACEBAD, 'fore:#00009D,back:#FF0000')
        # Indentation guide
        self.StyleSetSpec(stc.STC_STYLE_INDENTGUIDE, 'fore:#CDCDCD')
        # Python styles
        self.StyleSetSpec(stc.STC_P_DEFAULT, 'fore:#000000')
        # Comments
        self.StyleSetSpec(stc.STC_P_COMMENTLINE, 'fore:#008000,back:#F0FFF0')
        self.StyleSetSpec(stc.STC_P_COMMENTBLOCK, 'fore:#008000,back:#F0FFF0')
        # Numbers
        self.StyleSetSpec(stc.STC_P_NUMBER, 'fore:#008080')
        # Strings and characters
        self.StyleSetSpec(stc.STC_P_STRING, 'fore:#800080')
        self.StyleSetSpec(stc.STC_P_CHARACTER, 'fore:#800080')
        # Keywords
        self.StyleSetSpec(stc.STC_P_WORD, 'fore:#000080,bold')
        # Triple quotes
        self.StyleSetSpec(stc.STC_P_TRIPLE, 'fore:#800080,back:#FFFFEA')
        self.StyleSetSpec(stc.STC_P_TRIPLEDOUBLE, 'fore:#800080,back:#FFFFEA')
        # Class names
        self.StyleSetSpec(stc.STC_P_CLASSNAME, 'fore:#0000FF,bold')
        # Function names
        self.StyleSetSpec(stc.STC_P_DEFNAME, 'fore:#008080,bold')
        # Operators
        self.StyleSetSpec(stc.STC_P_OPERATOR, 'fore:#800000,bold')
        # Identifiers. I leave this as not bold because everything seems
        # to be an identifier if it doesn't match the above criterae
        self.StyleSetSpec(stc.STC_P_IDENTIFIER, 'fore:#000000')
        # Caret color
        self.SetCaretForeground('#007AFF')
        # Selection background
        self.SetSelBackground(1, '#66CCFF')
        self.SetSelBackground(
            True, wx.SystemSettings.GetColour(wx.SYS_COLOUR_HIGHLIGHT))
        self.SetSelForeground(
            True, wx.SystemSettings.GetColour(wx.SYS_COLOUR_HIGHLIGHTTEXT))
        self.SetWrapMode(stc.STC_WRAP_WORD)
        # indicator
        self.IndicatorSetStyle(0, stc.STC_INDIC_ROUNDBOX)
        self.IndicatorSetForeground(0, '#FF3B30')

    def prepandText(self, text):
        """Comment section"""
        doc = self
        sel = doc.GetSelection()
        start = doc.LineFromPosition(sel[0])
        end = doc.LineFromPosition(sel[1])
        end_pos = sel[1]
        if start > end:
            (start, end) = (end, start)
            end_pos = sel[0]
        if end > start and doc.GetColumn(end_pos) == 0:
            end = end - 1
        doc.BeginUndoAction()
        for line in six.moves.range(start, end + 1):
            firstChar = doc.PositionFromLine(line)
            doc.InsertText(firstChar, text)
        doc.SetCurrentPos(doc.PositionFromLine(start))
        doc.SetAnchor(doc.GetLineEndPosition(end))
        doc.EndUndoAction()
        doc.Refresh()

    def deprepandText(self, text):
        """Uncomment section"""
        doc = self
        sel = doc.GetSelection()
        start = doc.LineFromPosition(sel[0])
        end = doc.LineFromPosition(sel[1])
        end_pos = sel[1]
        if start > end:
            (start, end) = (end, start)
            end_pos = sel[0]
        if end > start and doc.GetColumn(end_pos) == 0:
            end = end - 1
        doc.BeginUndoAction()
        for line in six.moves.range(start, end + 1):
            firstChar = doc.PositionFromLine(line)
            txt = doc.GetLine(line)
            if txt.startswith(text):
                doc.SetCurrentPos(firstChar + len(text))
                doc.DelLineLeft()
        doc.SetSelection(sel[0], doc.PositionFromLine(end + 1))
        doc.SetCurrentPos(doc.PositionFromLine(start))
        doc.EndUndoAction()
        doc.Refresh()

    def indented(self):
        """increase the indent"""
        tabWidth = max(1, self.GetTabWidth())
        if self.GetUseTabs():
            indentation = '\t'
        else:
            indentation = ' ' * tabWidth
        self.prepandText(indentation)

    def unindented(self):
        """decrease the indent"""
        tabWidth = max(1, self.GetTabWidth())
        if self.GetUseTabs():
            indentation = '\t'
        else:
            indentation = ' ' * tabWidth
        self.deprepandText(indentation)

    def comment(self):
        """Comment section"""
        self.prepandText('##')

    def uncomment(self):
        """Uncomment section"""
        self.deprepandText('##')

    def GetContextMenu(self):
        """
            Create and return a context menu for the shell.
            This is used instead of the scintilla default menu
            in order to correctly respect our immutable buffer.
        """
        menu = wx.Menu()
        menu.Append(wx.ID_UNDO, 'Undo')
        menu.Append(wx.ID_REDO, 'Redo')
        menu.AppendSeparator()
        menu.Append(wx.ID_CUT, 'Cut')
        menu.Append(wx.ID_COPY, 'Copy')
        menu.Append(wx.ID_PASTE, 'Paste')
        menu.Append(wx.ID_CLEAR, 'Clear')
        menu.AppendSeparator()
        menu.Append(wx.ID_SELECTALL, 'Select All')
        menu.AppendSeparator()
        menu.Append(self.ID_COMMENT, 'Comment')
        menu.Append(self.ID_UNCOMMENT, 'Uncomment')
        return menu

    def OnContextMenu(self, evt):
        p = self.ScreenToClient(evt.GetPosition())
        m = self.GetMarginWidth(0) + self.GetMarginWidth(1)
        if p.x > m:
            # show edit menu when the mouse is in editable area
            menu = self.GetContextMenu()
            self.PopupMenu(menu)
        elif p.x > self.GetMarginWidth(0):
            # in breakpoint area
            cline = self.LineFromPosition(self.PositionFromPoint(p))
            for key in self.breakpointlist:
                line = self.MarkerLineFromHandle(key)
                if line == cline:
                    self.GotoLine(line)
                    break
            else:
                return
            menu = wx.Menu()
            menu.Append(self.ID_DELETE_BREAKPOINT, 'Delete Breakpoint')
            menu.AppendSeparator()
            menu.Append(self.ID_EDIT_BREAKPOINT, 'Condition...')
            menu.AppendSeparator()
            menu.Append(self.ID_CLEAR_BREAKPOINT, 'Delete All Breakpoints')
            self.PopupMenu(menu)

    def OnUpdateCommandUI(self, evt):
        eid = evt.Id
        if eid in (wx.ID_CUT, wx.ID_CLEAR):
            evt.Enable(self.GetSelectionStart() != self.GetSelectionEnd())
        elif eid == wx.ID_COPY:
            evt.Enable(self.GetSelectionStart() != self.GetSelectionEnd())
        elif eid == wx.ID_PASTE:
            evt.Enable(self.CanPaste())
        elif eid == wx.ID_UNDO:
            evt.Enable(self.CanUndo())
        elif eid == wx.ID_REDO:
            evt.Enable(self.CanRedo())
        else:
            evt.Skip()

    def findBreakPoint(self, line):
        for key in self.breakpointlist:
            if line == self.MarkerLineFromHandle(key):
                return self.breakpointlist[key]
        return None

    def OnProcessEvent(self, evt):
        """process the menu command"""
        eid = evt.GetId()
        if eid == wx.ID_CUT:
            self.Cut()
        elif eid == wx.ID_CLEAR:
            self.ClearAll()
        elif eid == wx.ID_COPY:
            self.Copy()
        elif eid == wx.ID_PASTE:
            self.Paste()
        elif eid == wx.ID_UNDO:
            self.Undo()
        elif eid == wx.ID_REDO:
            self.Redo()
        elif eid == wx.ID_SELECTALL:
            self.SelectAll()
        elif eid == self.ID_COMMENT:
            self.comment()
        elif eid == self.ID_UNCOMMENT:
            self.uncomment()
        elif eid == self.ID_DELETE_BREAKPOINT:
            bp = self.findBreakPoint(self.GetCurrentLine())
            if bp:
                dp.send('debugger.clear_breakpoint', id=bp['id'])
        elif eid == self.ID_CLEAR_BREAKPOINT:
            self.ClearBreakpoint()
        elif eid == self.ID_EDIT_BREAKPOINT:
            bp = self.findBreakPoint(self.GetCurrentLine())
            if bp:
                dlg = BreakpointSettingsDlg(self,
                                            bp['condition'], bp['hitcount'],
                                            bp.get('tcount', 0))
                if dlg.ShowModal() == wx.ID_OK:
                    cond = dlg.GetCondition()
                    dp.send('debugger.edit_breakpoint',
                            id=bp['id'],
                            condition=cond[0],
                            hitcount=cond[1])


class PyEditorPanel(wx.Panel):
    Gce = Gcm()
    ID_RUN_SCRIPT = wx.NewId()
    ID_DEBUG_SCRIPT = wx.NewId()
    ID_FIND_REPLACE = wx.NewId()
    ID_CHECK_SCRIPT = wx.NewId()
    ID_RUN_LINE = wx.NewId()
    ID_FIND_NEXT = wx.NewId()
    ID_FIND_PREV = wx.NewId()
    ID_INDENT = wx.NewId()
    ID_UNINDENT = wx.NewId()
    ID_SETCURFOLDER = wx.NewId()
    ID_TIDY_SOURCE = wx.NewId()
    ID_SPLIT_VERT = wx.NewId()
    ID_SPLIT_HORZ = wx.NewId()
    ID_DBG_RUN = wx.NewId()
    ID_DBG_STOP = wx.NewId()
    ID_DBG_STEP = wx.NewId()
    ID_DBG_STEP_INTO = wx.NewId()
    ID_DBG_STEP_OUT = wx.NewId()
    wildcard = 'Python source (*.py)|*.py|Text (*.txt)|*.txt|All files (*.*)|*.*'
    frame = None

    def __init__(self, parent):
        wx.Panel.__init__(self, parent, size=(1, 1))
        # find & replace dialog
        self.findStr = ""
        self.replaceStr = ""
        self.findFlags = 1
        self.stcFindFlags = 0
        self.wrapped = 0

        self.fileName = """"""
        self.splitter = wx.SplitterWindow(self, style=wx.SP_LIVE_UPDATE)
        self.editor = PyEditor(self.splitter)
        self.editor2 = None
        self.splitter.Initialize(self.editor)
        self.Bind(stc.EVT_STC_CHANGE, self.OnCodeModified)
        self.findDialog = None
        item = (
            (wx.ID_OPEN, 'Open', open_xpm, 'Open Python script'),
            (wx.ID_REFRESH, 'Reload', reload_xpm, 'Reload current script'),
            (wx.ID_SAVE, 'Save', save_xpm, 'Save current document (Ctrl+S)'),
            (wx.ID_SAVEAS, 'Save As', saveas_xpm, 'Save current document as'),
            (None, None, None, None),
            (self.ID_FIND_REPLACE, 'Find', find_xpm, 'Find/Replace (Ctrl+F)'),
            (None, None, None, None),
            (self.ID_INDENT, 'Increase Indent', indent_xpm,
             'Increase the indent'),
            (self.ID_UNINDENT, 'Decrease Indent', dedent_xpm,
             'Decrease the indent'),
            (None, None, None, None),
            (self.ID_RUN_LINE, 'Run', run_xpm,
             'Run the current line or selection (Ctrl+Return)'),
            (self.ID_RUN_SCRIPT, 'Execute', execute_xpm,
             'Execute the whole script'),
            (None, None, None, None),
            (self.ID_CHECK_SCRIPT, 'Check', check_xpm, 'Check the module'),
            (self.ID_DEBUG_SCRIPT, 'Debug', debug_xpm, 'Debug the script'),
            (None, None, None, None),
            (self.ID_SETCURFOLDER, 'Set current folder', folder_xpm,
             'Set the file folder as current folder'),
            (None, None, None, None),
            (self.ID_SPLIT_VERT, 'Split Vert', vert_xpm,
             'Split the window vertically'),
            (self.ID_SPLIT_HORZ, 'Split Horz', horz_xpm,
             'Split the window horizontally'),
        )

        self.toolbarart = AuiToolBarPopupArt(self)
        self.tb = aui.AuiToolBar(self,
                                 agwStyle=aui.AUI_TB_OVERFLOW
                                 | aui.AUI_TB_PLAIN_BACKGROUND)
        for (eid, label, img_xpm, tooltip) in item:
            if eid == None:
                self.tb.AddSeparator()
                continue
            bmp = wx.Bitmap(to_byte(img_xpm))
            if label in ['Split Vert', 'Split Horz']:
                self.tb.AddCheckTool(eid, label, bmp, wx.NullBitmap, tooltip)
            else:
                self.tb.AddSimpleTool(eid, label, bmp, tooltip)
        self.tb.AddSeparator()
        self.cbWrapMode = wx.CheckBox(self.tb, wx.ID_ANY, 'Word Wrap')
        self.cbWrapMode.SetValue(True)
        self.tb.AddControl(self.cbWrapMode)

        self.tb.SetArtProvider(self.toolbarart)
        self.tb.Realize()
        self.box = wx.BoxSizer(wx.VERTICAL)
        self.box.Add(self.tb, 0, wx.EXPAND, 5)
        self.box.Add(wx.StaticLine(self), 0, wx.EXPAND)
        self.box.Add(self.splitter, 1, wx.EXPAND)
        self.box.Fit(self)
        self.SetSizer(self.box)
        # Connect Events
        self.Bind(wx.EVT_TOOL, self.OnBtnOpen, id=wx.ID_OPEN)
        self.Bind(wx.EVT_TOOL, self.OnBtnReload, id=wx.ID_REFRESH)
        self.Bind(wx.EVT_TOOL, self.OnBtnSave, id=wx.ID_SAVE)
        self.Bind(wx.EVT_TOOL, self.OnBtnSaveAs, id=wx.ID_SAVEAS)
        self.tb.Bind(wx.EVT_UPDATE_UI, self.OnUpdateBtn)
        self.Bind(wx.EVT_TOOL, self.OnShowFindReplace, id=self.ID_FIND_REPLACE)
        self.Bind(wx.EVT_TOOL, self.OnBtnRun, id=self.ID_RUN_LINE)
        self.Bind(wx.EVT_TOOL, self.OnBtnCheck, id=self.ID_CHECK_SCRIPT)
        self.Bind(wx.EVT_TOOL, self.OnBtnRunScript, id=self.ID_RUN_SCRIPT)
        self.Bind(wx.EVT_TOOL, self.OnBtnDebugScript, id=self.ID_DEBUG_SCRIPT)
        #self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateBtn, id=self.ID_DEBUG_SCRIPT)
        self.Bind(wx.EVT_TOOL, self.OnFindNext, id=self.ID_FIND_NEXT)
        self.Bind(wx.EVT_TOOL, self.OnFindPrev, id=self.ID_FIND_PREV)
        self.Bind(wx.EVT_TOOL, self.OnIndent, id=self.ID_INDENT)
        self.Bind(wx.EVT_TOOL, self.OnUnindent, id=self.ID_UNINDENT)
        self.Bind(wx.EVT_TOOL, self.OnSetCurFolder, id=self.ID_SETCURFOLDER)
        self.Bind(wx.EVT_TOOL, self.OnSplitVert, id=self.ID_SPLIT_VERT)
        self.Bind(wx.EVT_TOOL, self.OnSplitHorz, id=self.ID_SPLIT_HORZ)
        self.cbWrapMode.Bind(wx.EVT_CHECKBOX, self.OnWrap)
        accel = [
            (wx.ACCEL_CTRL, ord('F'), self.ID_FIND_REPLACE),
            (wx.ACCEL_NORMAL, wx.WXK_F3, self.ID_FIND_NEXT),
            (wx.ACCEL_SHIFT, wx.WXK_F3, self.ID_FIND_PREV),
            (wx.ACCEL_CTRL, ord('H'), self.ID_FIND_REPLACE),
            (wx.ACCEL_CTRL, wx.WXK_RETURN, self.ID_RUN_LINE),
            (wx.ACCEL_CTRL, ord('S'), wx.ID_SAVE),
        ]
        self.accel = wx.AcceleratorTable(accel)
        self.SetAcceleratorTable(self.accel)
        #dp.connect(self.debug_paused, 'debugger.paused')
        dp.connect(self.debug_ended, 'debugger.ended')
        dp.connect(self.debug_bpadded, 'debugger.breakpoint_added')
        dp.connect(self.debug_bpcleared, 'debugger.breakpoint_cleared')
        self.debug_curline = None
        self.num = self.Gce.get_next_num()
        self.Gce.set_active(self)

    @classmethod
    def get_instances(cls):
        for inst in cls.Gce.get_all_managers():
            yield inst

    def Destroy(self):
        """destroy the panel"""
        self.editor.ClearBreakpoint()
        self.CheckModified()
        self.Gce.destroy(self.num)
        return super(PyEditorPanel, self).Destroy()

    def update_bp(self):
        """update the breakpoints"""
        for key in self.editor.breakpointlist:
            line = self.editor.MarkerLineFromHandle(key) + 1
            if line != self.editor.breakpointlist[key]['lineno']:
                ids = self.editor.breakpointlist[key]['id']
                dp.send('debugger.edit_breakpoint', id=ids, lineno=line)

    def debug_bpadded(self, bpdata):
        """the breakpoint is added"""
        if bpdata is None:
            return
        info = bpdata
        filename = info['filename']
        if filename != self.editor.filename:
            return
        for key in self.editor.breakpointlist:
            if self.editor.breakpointlist[key]['id'] == bpdata['id']:
                return
        lineno = info['lineno']
        handler = self.editor.MarkerAdd(lineno - 1, 0)
        self.editor.breakpointlist[handler] = bpdata

    def debug_bpcleared(self, bpdata):
        """the breakpoint is cleared"""
        if bpdata is None:
            return
        info = bpdata
        filename = info['filename']
        if filename != self.editor.filename:
            return
        for key in self.editor.breakpointlist:
            if self.editor.breakpointlist[key]['id'] == bpdata['id']:
                self.editor.MarkerDeleteHandle(key)
                del self.editor.breakpointlist[key]
                break

    def debug_paused(self, status):
        """the debug is paused"""
        # delete the current line marker
        if self.debug_curline:
            self.editor.MarkerDeleteHandle(self.debug_curline)
            self.debug_curline = None
        if status is None:
            return False
        filename = status['filename']

        lineno = -1
        marker = -1
        active = False
        if filename == self.editor.filename:
            lineno = status['lineno']
            marker = 1
            active = True
        else:
            frames = status['frames']
            if frames is not None:
                for frame in frames:
                    filename = inspect.getsourcefile(frame) or inspect.getfile(
                        frame)
                    if filename == self.fileName:
                        lineno = frame.f_lineno
                        marker = 2
                        break
        if lineno >= 0 and marker >= 0:
            self.debug_curline = self.editor.MarkerAdd(lineno - 1, marker)
            self.editor.EnsureVisibleEnforcePolicy(lineno - 1)
            #self.JumpToLine(lineno-1)
            #self.editor.GotoLine(lineno-1)
            #self.editor.EnsureVisible(lineno-1)
            #self.editor.EnsureCaretVisible()

            if active:
                show = self.IsShown()
                parent = self.GetParent()
                while show and parent:
                    show = parent.IsShown()
                    parent = parent.GetParent()
                if not show:
                    dp.send('frame.show_panel', panel=self)
            return True
        return False

    def debug_ended(self):
        """debugging finished"""
        if self.debug_curline:
            # hide the marker
            self.editor.MarkerDeleteHandle(self.debug_curline)
            self.debug_curline = None

    def OnWrap(self, event):
        """turn on/off the wrap mode"""
        if self.cbWrapMode.IsChecked():
            self.editor.SetWrapMode(stc.STC_WRAP_WORD)
        else:
            self.editor.SetWrapMode(stc.STC_WRAP_NONE)

    def JumpToLine(self, line, highlight=False):
        """jump to the line and make sure it is visible"""
        self.editor.GotoLine(line)
        self.editor.SetFocus()
        if highlight:
            self.editor.SelectLine(line)
        wx.CallLater(1, self.editor.EnsureCaretVisible)

    def OnCodeModified(self, event):
        """called when the file is modified"""
        filename = 'untiled'
        if self.fileName != "":
            (_, filename) = os.path.split(self.fileName)
        if self.editor.GetModify():
            filename = filename + '*'
        dp.send('frame.set_panel_title', pane=self, title=filename)

    def LoadFile(self, path):
        """open file"""
        self.editor.LoadFile(path)
        self.fileName = path
        (_, filename) = os.path.split(self.fileName)
        dp.send('frame.set_panel_title', pane=self, title=filename)

    def OnBtnOpen(self, event):
        """open the script"""
        defaultDir = os.path.dirname(self.fileName)
        style = wx.FD_OPEN | wx.FD_FILE_MUST_EXIST
        dlg = wx.FileDialog(self,
                            'Open',
                            defaultDir=defaultDir,
                            wildcard=self.wildcard,
                            style=style)
        if dlg.ShowModal() == wx.ID_OK:
            path = dlg.GetPaths()[0]
            self.LoadFile(path)
        dlg.Destroy()

    def OnBtnReload(self, event):
        """reload file"""
        if self.fileName:
            self.LoadFile(self.fileName)

    def saveFile(self):
        if self.fileName == "":
            defaultDir = os.path.dirname(self.fileName)
            # use top level frame as parent, otherwise it may crash when
            # it is called in Destroy()
            style = wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT | wx.FD_CHANGE_DIR
            dlg = wx.FileDialog(self.GetTopLevelParent(),
                                'Save As',
                                defaultDir=defaultDir,
                                wildcard=self.wildcard,
                                style=style)
            if dlg.ShowModal() == wx.ID_OK:
                path = dlg.GetPath()
                self.fileName = path
            dlg.Destroy()
        self.editor.SaveFile(self.fileName)
        (path, filename) = os.path.split(self.fileName)
        dp.send('frame.set_panel_title', pane=self, title=filename)
        self.update_bp()

    def OnBtnSave(self, event):
        """save the script"""
        self.saveFile()

    def OnBtnSaveAs(self, event):
        """save the script with different filename"""
        defaultDir = os.path.dirname(self.fileName)
        style = wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT | wx.FD_CHANGE_DIR
        dlg = wx.FileDialog(self,
                            'Save As',
                            defaultDir=defaultDir,
                            wildcard=self.wildcard,
                            style=style)
        if dlg.ShowModal() == wx.ID_OK:
            path = dlg.GetPaths()[0]
            self.fileName = path
            dlg.Destroy()
        self.editor.SaveFile(self.fileName)
        (path, filename) = os.path.split(self.fileName)
        dp.send('frame.set_panel_title', pane=self, title=filename)
        self.update_bp()

    def OnUpdateBtn(self, event):
        """update the toolbar button status"""
        eid = event.GetId()
        if eid == wx.ID_SAVE:
            event.Enable(self.editor.GetModify())
        elif eid == self.ID_DEBUG_SCRIPT:
            resp = dp.send('debugger.debugging')
            if resp:
                event.Enable(not resp[0][1])
        elif eid == wx.ID_REFRESH:
            event.Enable(self.fileName != "")

    def OnShowFindReplace(self, event):
        """Find and Replace dialog and action."""
        # find string
        findStr = self.editor.GetSelectedText()
        if findStr and self.findDialog:
            self.findDialog.Destroy()
            self.findDialog = None
        # dialog already open, if yes give focus
        if self.findDialog:
            self.findDialog.Show(1)
            self.findDialog.Raise()
            return
        if not findStr:
            findStr = self.findStr
        # find data
        data = wx.FindReplaceData(self.findFlags)
        data.SetFindString(findStr)
        data.SetReplaceString(self.replaceStr)
        # dialog
        self.findDialog = wx.FindReplaceDialog(
            self, data, 'Find & Replace', wx.FR_REPLACEDIALOG)
        # bind the event to the dialog, see the example in wxPython demo
        self.findDialog.Bind(wx.EVT_FIND, self.OnFind)
        self.findDialog.Bind(wx.EVT_FIND_NEXT, self.OnFind)
        self.findDialog.Bind(wx.EVT_FIND_REPLACE, self.OnReplace)
        self.findDialog.Bind(wx.EVT_FIND_REPLACE_ALL, self.OnReplaceAll)
        self.findDialog.Bind(wx.EVT_FIND_CLOSE, self.OnFindClose)
        self.findDialog.Show(1)
        self.findDialog.data = data  # save a reference to it...

    def RunCommand(self, command, prompt=False, verbose=True, debug=False):
        """run command in shell"""
        dp.send('shell.run',
                command=command,
                prompt=prompt,
                verbose=verbose,
                debug=debug)

    def OnBtnRun(self, event):
        """execute the selection or current line"""
        cmd = self.editor.GetSelectedText()
        if not cmd or cmd == """""":
            (cmd, _) = self.editor.GetCurLine()
            cmd = cmd.rstrip()
        lines = cmd.split('\n')
        for line in lines:
            self.RunCommand(line, prompt=True, verbose=True)

    def CheckModified(self):
        """check whether it is modified"""
        if self.editor.GetModify():
            msg = 'The file has been modified. Save it first?'
            # use top level frame as parent, otherwise it may crash when
            # it is called in Destroy()
            dlg = wx.MessageDialog(self.GetTopLevelParent(), msg, 'bsmedit',
                                   wx.YES_NO)
            result = dlg.ShowModal() == wx.ID_YES
            dlg.Destroy()
            if result:
                self.saveFile()
            return self.editor.GetModify()
        return False

    def OnBtnCheck(self, event):
        """check the syntax"""
        if self.CheckModified():
            return
        if self.fileName == """""":
            return
        self.RunCommand('import sys', verbose=False)
        self.RunCommand('_bsm_source = open(r\'%s\',\'r\').read()+\'\\n\'' %
                        self.fileName,
                        verbose=False)
        self.RunCommand('compile(_bsm_source,r\'%s\',\'exec\')' %
                        self.fileName,
                        prompt=True,
                        verbose=True)
        self.RunCommand('del _bsm_source', verbose=False)

    def OnBtnRunScript(self, event):
        """execute the script"""
        if self.CheckModified():
            return
        if not self.fileName:
            return
        self.RunCommand('import six', verbose=False)
        cmd = "compile(open(r'{0}', 'rb').read(), r'{0}', 'exec')".format(
            self.fileName)
        self.RunCommand('six.exec_(%s)' % cmd,
                        prompt=True,
                        verbose=False,
                        debug=False)

    def OnBtnDebugScript(self, event):
        """execute the script in debug mode"""
        if self.CheckModified():
            return
        if not self.fileName:
            return
        self.RunCommand('import six', verbose=False)
        # disable the debugger button
        self.tb.EnableTool(self.ID_DEBUG_SCRIPT, False)

        cmd = "compile(open(r'{0}', 'rb').read(), r'{0}', 'exec')".format(
            self.fileName)
        self.RunCommand('six.exec_(%s)' % cmd,
                        prompt=True,
                        verbose=False,
                        debug=True)

        #dp.send('debugger.ended')
        self.tb.EnableTool(self.ID_DEBUG_SCRIPT, True)

    def message(self, text):
        """show the message on statusbar"""
        dp.send('frame.show_status_text', text=text)

    def _find_text(self, minPos, maxPos, text, flags=0):
        position = self.editor.FindText(minPos, maxPos, text, flags)
        if isinstance(position, tuple):
            position = position[0] # wx ver 4.1.0 returns (start, end)
        return position

    def doFind(self, strFind, forward=True):
        """search the string"""
        current = self.editor.GetCurrentPos()
        position = -1
        if forward:
            position = self._find_text(current, len(self.editor.GetText()),
                                       strFind, self.stcFindFlags)
            if position == -1:
                # wrap around
                self.wrapped += 1
                position = self._find_text(0, current + len(strFind), strFind,
                                           self.stcFindFlags)
        else:
            position = self._find_text(current - len(strFind), 0, strFind,
                                       self.stcFindFlags)
            if position == -1:
                # wrap around
                self.wrapped += 1
                position = self._find_text(len(self.editor.GetText()), current,
                                           strFind, self.stcFindFlags)

        # not found the target, do not change the current position
        if position == -1:
            self.message("'%s' not found!" % strFind)
            position = current
            strFind = """"""
        self.editor.GotoPos(position)
        self.editor.SetSelection(position, position + len(strFind))
        return position

    def OnFind(self, event):
        """search the string"""
        self.findStr = event.GetFindString()
        self.findFlags = event.GetFlags()
        flags = 0
        if wx.FR_WHOLEWORD & self.findFlags:
            flags |= stc.STC_FIND_WHOLEWORD
        if wx.FR_MATCHCASE & self.findFlags:
            flags |= stc.STC_FIND_MATCHCASE
        self.stcFindFlags = flags
        return self.doFind(self.findStr, wx.FR_DOWN & self.findFlags)

    def OnFindClose(self, event):
        """close find & replace dialog"""
        event.GetDialog().Destroy()

    def OnReplace(self, event):
        """replace"""
        # Next line avoid infinite loop
        findStr = event.GetFindString()
        self.replaceStr = event.GetReplaceString()

        source = self.editor
        selection = source.GetSelectedText()
        if not event.GetFlags() & wx.FR_MATCHCASE:
            findStr = findStr.lower()
            selection = selection.lower()

        if selection == findStr:
            position = source.GetSelectionStart()
            source.ReplaceSelection(self.replaceStr)
            source.SetSelection(position, position + len(self.replaceStr))
        # jump to next instance
        position = self.OnFind(event)
        return position

    def OnReplaceAll(self, event):
        """replace all the instances"""
        source = self.editor
        count = 0
        self.wrapped = 0
        position = start = source.GetCurrentPos()
        while position > -1 and (not self.wrapped or position < start):
            position = self.OnReplace(event)
            if position != -1:
                count += 1
            if self.wrapped >= 2:
                break
        self.editor.GotoPos(start)
        if not count:
            self.message("'%s' not found!" % event.GetFindString())

    def OnFindNext(self, event):
        """go the previous instance of search string"""
        findStr = self.editor.GetSelectedText()
        if findStr:
            self.findStr = findStr
        if self.findStr:
            self.doFind(self.findStr)

    def OnFindPrev(self, event):
        """go the previous instance of search string"""
        findStr = self.editor.GetSelectedText()
        if findStr:
            self.findStr = findStr
        if self.findStr:
            self.doFind(self.findStr, False)

    def OnIndent(self, event):
        """increase the indent"""
        self.editor.indented()

    def OnUnindent(self, event):
        """decrease the indent"""
        self.editor.unindented()

    def OnSetCurFolder(self, event):
        """set the current folder to the folder with the file"""
        if not self.fileName:
            return
        path, = os.path.split(self.fileName)
        self.RunCommand('import os', verbose=False)
        self.RunCommand('os.chdir(r\'%s\')' % path, verbose=False)

    def OnSplitVert(self, event):
        """show splitter window vertically"""
        show = self.tb.GetToolState(self.ID_SPLIT_VERT)
        if not show:
            # hide the splitter window
            if self.editor2:
                if self.splitter.IsSplit():
                    self.splitter.Unsplit(self.editor2)
                self.editor2.Hide()
        else:
            # show splitter window
            if not self.editor2:
                # create the splitter window
                self.editor2 = PyEditor(self.splitter)
                self.editor2.SetDocPointer(self.editor.GetDocPointer())
            if self.editor2:
                if self.splitter.IsSplit():
                    self.splitter.Unsplit(self.editor2)
                self.splitter.SplitHorizontally(self.editor, self.editor2)
                self.tb.ToggleTool(self.ID_SPLIT_HORZ, False)

    def OnSplitHorz(self, event):
        """show splitter window horizontally"""
        show = self.tb.GetToolState(self.ID_SPLIT_HORZ)
        if not show:
            # hide the splitter window
            if self.editor2:
                if self.splitter.IsSplit():
                    self.splitter.Unsplit(self.editor2)
                self.editor2.Hide()
        else:
            # show splitter window
            if not self.editor2:
                # create the splitter window
                self.editor2 = PyEditor(self.splitter)
                self.editor2.SetDocPointer(self.editor.GetDocPointer())
            if self.editor2:
                if self.splitter.IsSplit():
                    self.splitter.Unsplit(self.editor2)
                self.splitter.SplitVertically(self.editor, self.editor2)
                self.tb.ToggleTool(self.ID_SPLIT_VERT, False)

    @classmethod
    def Initialize(cls, frame, **kwargs):
        """initialize the module"""
        if cls.frame:
            # if it has already initialized, simply return
            return
        cls.frame = frame
        cls.kwargs = kwargs
        resp = dp.send('frame.add_menu',
                       path='File:New:Python script\tCtrl+N',
                       rxsignal='bsm.editor.menu')
        if resp:
            cls.ID_EDITOR_NEW = resp[0][1]
        resp = dp.send('frame.add_menu',
                       path='File:Open:Python script\tctrl+O',
                       rxsignal='bsm.editor.menu')
        if resp:
            cls.ID_EDITOR_OPEN = resp[0][1]
        dp.connect(cls.ProcessCommand, 'bsm.editor.menu')
        dp.connect(cls.Uninitialize, 'frame.exit')
        dp.connect(cls.OpenScript, 'frame.file_drop')
        dp.connect(cls.DebugPaused, 'debugger.paused')
        dp.connect(cls.DebugUpdateScope, 'debugger.update_scopes')

    @classmethod
    def DebugPaused(cls):
        """the debugger has paused, update the editor margin marker"""
        resp = dp.send('debugger.get_status')
        if not resp or not resp[0][1]:
            return
        status = resp[0][1]
        filename = status['filename']
        # open the file if necessary
        editor = cls.OpenScript(filename)
        if editor:
            editor.debug_paused(status)
        for editor2 in PyEditorPanel.get_instances():
            if editor != editor2:
                editor2.debug_paused(status)

    @classmethod
    def DebugUpdateScope(cls):
        """
        the debugger scope has been changed, update the editor margin marker
        """
        resp = dp.send('debugger.get_status')
        if not resp or not resp[0][1]:
            return
        status = resp[0][1]
        for editor in PyEditorPanel.get_instances():
            editor.debug_paused(status)

    @classmethod
    def Uninitialize(cls):
        """unload the module"""
        pass

    @classmethod
    def ProcessCommand(cls, command):
        """process the menu command"""
        if command == cls.ID_EDITOR_NEW:
            cls.AddEditor()
        elif command == cls.ID_EDITOR_OPEN:
            defaultDir = os.path.dirname(os.getcwd())

            style = wx.FD_OPEN | wx.FD_FILE_MUST_EXIST
            dlg = wx.FileDialog(cls.frame,
                                'Open',
                                defaultDir=defaultDir,
                                wildcard=cls.wildcard,
                                style=style)
            if dlg.ShowModal() == wx.ID_OK:
                path = dlg.GetPaths()[0]
                cls.OpenScript(path)
            dlg.Destroy()

    @classmethod
    def AddEditor(cls, title='untitle', activated=True):
        """create a editor panel"""
        editor = PyEditorPanel(cls.frame)

        direction = cls.kwargs.get('direction', 'top')
        dp.send("frame.add_panel",
                panel=editor,
                title=title,
                active=activated,
                direction=direction)
        return editor

    @classmethod
    def OpenScript(cls, filename, activated=True, lineno=0):
        """open the file"""
        if not filename:
            return None
        (_, fileExtension) = os.path.splitext(filename)
        if fileExtension.lower() != '.py':
            return None

        editor = cls.findEditorByFileName(filename)
        if editor is None:
            editor = cls.AddEditor()
            editor.LoadFile(filename)
            dp.send('frame.add_file_history', filename=filename)

        if editor and activated and not editor.IsShown():
            dp.send('frame.show_panel', panel=editor, focus=True)
        if lineno > 0:
            editor.JumpToLine(lineno - 1, True)
        return editor

    @classmethod
    def findEditorByFileName(cls, filename):
        """
        find the editor by filename

        If the file is opened in multiple editors, return the first one.
        """
        for editor in PyEditorPanel.get_instances():
            if str(editor.editor.filename).lower() == filename.lower():
                return editor
        return None


def bsm_initialize(frame, **kwargs):
    """initialize the model"""
    PyEditorPanel.Initialize(frame, **kwargs)



"""copy the docstring"""
import inspect
import textwrap


def copy(source, dropSelf=True):
    """return the function to copy the docstring"""
    def do_copy(target):
        """do the real copy"""
        if source.__doc__:
            argspec = inspect.formatargspec(*inspect.getargspec(source))
            if dropSelf:
                # The first parameter to a method is a reference to an
                # instance, usually coded as "self", and is usually passed
                # automatically by Python; therefore we want to drop it.
                temp = argspec.split(',')
                if len(temp) == 1:  # No other arguments.
                    argspec = '()'
                elif temp[
                        0][:2] == '(*':  # first param is like *args, not self
                    pass
                else:  # Drop the first argument.
                    argspec = '(' + ','.join(temp[1:]).lstrip()

            docstring = textwrap.dedent(source.__doc__)
            if target.__doc__:
                docstring += textwrap.dedent(target.__doc__)

            target.__doc__ = docstring
        return target

    return do_copy


def copy_docstring(source, dropSelf=True):
    """copy the docstring to target, used as decorator"""
    return lambda target: copy(source, dropSelf)(target)


def copy_docstring_raw(source, target, dropSelf=True):
    """copy the docstring to target"""
    if source.__doc__:
        argspec = inspect.formatargspec(*inspect.getargspec(source))
        if dropSelf:
            # The first parameter to a method is a reference to an
            # instance, usually coded as "self", and is usually passed
            # automatically by Python; therefore we want to drop it.
            temp = argspec.split(',')
            if len(temp) == 1:  # No other arguments.
                argspec = '()'
            elif temp[0][:2] == '(*':  # first param is like *args, not self
                pass
            else:  # Drop the first argument.
                argspec = '(' + ','.join(temp[1:]).lstrip()
        docstring = '\n' + source.__name__ + argspec + '\n'
        docstring += '\n' + textwrap.dedent(source.__doc__).strip() + '\n'

        target.__doc__ = docstring
    return target



from matplotlib.figure import Figure
from matplotlib._pylab_helpers import Gcf
from matplotlib.backends.backend_wx import Show
from matplotlib import is_interactive
from matplotlib import get_backend
from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg as FigureCanvas


def new_figure_manager(num, *args, **kwargs):
    """
    Create a new figure manager instance
    """

    # in order to expose the Figure constructor to the pylab
    # interface we need to create the figure here

    from .graph import MatplotPanel
    FigureClass = kwargs.pop('FigureClass', Figure)
    thisFig = FigureClass(*args, **kwargs)

    return MatplotPanel.addFigure('Figure %d' % num, num, thisFig)


def draw_if_interactive():
    """
    This should be overridden in a windowing environment if drawing
    should be done in interactive python mode
    """

    if is_interactive():
        figManager = Gcf.get_active()
        if figManager is not None:
            figManager.canvas.draw()


class ShowFigure(Show):
    def __call__(self, fig=None, block=None):
        """
        Show all figures.  If *block* is not None, then
        it is a boolean that overrides all other factors
        determining whether show blocks by calling mainloop().
        The other factors are:
        it does not block if run inside "ipython --pylab";
        it does not block in interactive mode.
        """

        if isinstance(fig, int):
            manager = Gcf.get_fig_manager(fig)
        else:
            manager = Gcf.get_active()
        if not manager:
            return

        # for manager in managers:

        manager.show()

        if block is not None:
            if block:
                self.mainloop()
                return
            else:
                return

        if not is_interactive() or get_backend() == 'WebAgg':
            self.mainloop()


show = ShowFigure()
